import os
import platform

import pytest
from uqbar.strings import normalize

from supriya.assets.synthdefs import system_link_audio_1, system_link_audio_2
from supriya.ugens import SuperColliderSynthDef, decompile_synthdef


def test_system_link_audio_1_supriya():
    assert normalize(str(system_link_audio_1)) == normalize(
        """
        synthdef:
            name: system_link_audio_1
            ugens:
            -   Control.kr:
                    done_action: 2.0
                    fade_time: 0.02
                    gate: 1.0
                    in_: 16.0
                    out: 0.0
            -   BinaryOpUGen(LESS_THAN_OR_EQUAL).kr:
                    left: Control.kr[1:fade_time]
                    right: 0.0
            -   EnvGen.kr:
                    gate: Control.kr[2:gate]
                    level_scale: 1.0
                    level_bias: 0.0
                    time_scale: Control.kr[1:fade_time]
                    done_action: Control.kr[0:done_action]
                    envelope[0]: BinaryOpUGen(LESS_THAN_OR_EQUAL).kr[0]
                    envelope[1]: 2.0
                    envelope[2]: 1.0
                    envelope[3]: -99.0
                    envelope[4]: 1.0
                    envelope[5]: 1.0
                    envelope[6]: 3.0
                    envelope[7]: 0.0
                    envelope[8]: 0.0
                    envelope[9]: 1.0
                    envelope[10]: 3.0
                    envelope[11]: 0.0
            -   InFeedback.ar:
                    bus: Control.kr[3:in_]
            -   BinaryOpUGen(MULTIPLICATION).ar:
                    left: InFeedback.ar[0]
                    right: EnvGen.kr[0]
            -   Out.ar:
                    bus: Control.kr[4:out]
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
        """
    )


def test_system_link_audio_1_bytes():
    py_compiled = system_link_audio_1.compile()
    assert py_compiled == (
        b"SCgf\x00\x00\x00\x02\x00\x01\x13system_link_audio_1\x00\x00\x00\x05\x00\x00"
        b"\x00\x00?\x80\x00\x00@\x00\x00\x00\xc2\xc6\x00\x00@@\x00\x00\x00\x00"
        b"\x00\x05@\x00\x00\x00<\xa3\xd7\n?\x80\x00\x00A\x80\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x05\x0bdone_action\x00\x00\x00\x00\tfade_time"
        b"\x00\x00\x00\x01\x04gate\x00\x00\x00\x02\x03in_\x00\x00\x00\x03\x03ou"
        b"t\x00\x00\x00\x04\x00\x00\x00\x06\x07Control\x01\x00\x00\x00\x00\x00\x00"
        b"\x00\x05\x00\x00\x01\x01\x01\x01\x01\x0cBinaryOpUGen\x01\x00\x00\x00\x02\x00"
        b"\x00\x00\x01\x00\n\x00\x00\x00\x00\x00\x00\x00\x01\xff\xff\xff"
        b"\xff\x00\x00\x00\x00\x01\x06EnvGen\x01\x00\x00\x00\x11\x00\x00"
        b"\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\xff\xff\xff\xff"
        b"\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        b"\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x02\xff\xff\xff\xff"
        b"\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x03\xff\xff\xff\xff"
        b"\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x01\xff\xff\xff\xff"
        b"\x00\x00\x00\x04\xff\xff\xff\xff\x00\x00\x00\x00\xff\xff\xff\xff"
        b"\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x01\xff\xff\xff\xff"
        b"\x00\x00\x00\x04\xff\xff\xff\xff\x00\x00\x00\x00\x01\nInFeedback"
        b"\x02\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x03\x02\x0cBinaryOpUGen\x02\x00\x00\x00\x02\x00\x00\x00\x01\x00\x02"
        b"\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x02\x03Ou"
        b"t\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00"
    )


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_system_link_audio_1_sclang():
    sc_compiled = SuperColliderSynthDef(
        "system_link_audio_1",
        r"""
        | out=0, in=16, vol=1, level=1, lag=0.05, doneAction=2 |
        var env = EnvGate(i_level: 0, doneAction:doneAction, curve:'sin')
        * Lag.kr(vol * level, lag);
        Out.ar(out, InFeedback.ar(in, 1) * env)
        """,
        r"[\kr, \kr, \kr, \kr, \kr, \ir]",
    ).compile()
    sc_synthdef = decompile_synthdef(sc_compiled)
    assert normalize(str(sc_synthdef)) == normalize(
        """
        synthdef:
            name: system_link_audio_1
            ugens:
            -   Control.ir:
                    doneAction: 2.0
            -   Control.kr:
                    out: 0.0
                    in: 16.0
                    vol: 1.0
                    level: 1.0
                    lag: 0.05000000074505806
            -   BinaryOpUGen(MULTIPLICATION).kr/0:
                    left: Control.kr[2:vol]
                    right: Control.kr[3:level]
            -   Lag.kr:
                    source: BinaryOpUGen(MULTIPLICATION).kr/0[0]
                    lag_time: Control.kr[4:lag]
            -   InFeedback.ar:
                    bus: Control.kr[1:in]
            -   Control.kr:
                    gate: 1.0
            -   Control.kr:
                    fadeTime: 0.019999999552965164
            -   EnvGen.kr:
                    gate: Control.kr[0:gate]
                    level_scale: 1.0
                    level_bias: 0.0
                    time_scale: Control.kr[0:fadeTime]
                    done_action: Control.ir[0:doneAction]
                    envelope[0]: 0.0
                    envelope[1]: 2.0
                    envelope[2]: 1.0
                    envelope[3]: -99.0
                    envelope[4]: 1.0
                    envelope[5]: 1.0
                    envelope[6]: 3.0
                    envelope[7]: 0.0
                    envelope[8]: 0.0
                    envelope[9]: 1.0
                    envelope[10]: 3.0
                    envelope[11]: 0.0
            -   BinaryOpUGen(MULTIPLICATION).kr/1:
                    left: EnvGen.kr[0]
                    right: Lag.kr[0]
            -   BinaryOpUGen(MULTIPLICATION).ar:
                    left: InFeedback.ar[0]
                    right: BinaryOpUGen(MULTIPLICATION).kr/1[0]
            -   Out.ar:
                    bus: Control.kr[0:out]
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
        """
    )


def test_system_link_audio_2_supriya():
    assert normalize(str(system_link_audio_2)) == normalize(
        """
        synthdef:
            name: system_link_audio_2
            ugens:
            -   Control.kr:
                    done_action: 2.0
                    fade_time: 0.02
                    gate: 1.0
                    in_: 16.0
                    out: 0.0
            -   BinaryOpUGen(LESS_THAN_OR_EQUAL).kr:
                    left: Control.kr[1:fade_time]
                    right: 0.0
            -   EnvGen.kr:
                    gate: Control.kr[2:gate]
                    level_scale: 1.0
                    level_bias: 0.0
                    time_scale: Control.kr[1:fade_time]
                    done_action: Control.kr[0:done_action]
                    envelope[0]: BinaryOpUGen(LESS_THAN_OR_EQUAL).kr[0]
                    envelope[1]: 2.0
                    envelope[2]: 1.0
                    envelope[3]: -99.0
                    envelope[4]: 1.0
                    envelope[5]: 1.0
                    envelope[6]: 3.0
                    envelope[7]: 0.0
                    envelope[8]: 0.0
                    envelope[9]: 1.0
                    envelope[10]: 3.0
                    envelope[11]: 0.0
            -   InFeedback.ar:
                    bus: Control.kr[3:in_]
            -   BinaryOpUGen(MULTIPLICATION).ar/0:
                    left: InFeedback.ar[0]
                    right: EnvGen.kr[0]
            -   BinaryOpUGen(MULTIPLICATION).ar/1:
                    left: InFeedback.ar[1]
                    right: EnvGen.kr[0]
            -   Out.ar:
                    bus: Control.kr[4:out]
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                    source[1]: BinaryOpUGen(MULTIPLICATION).ar/1[0]
        """
    )


def test_system_link_audio_2_bytes():
    py_compiled = system_link_audio_2.compile()
    assert py_compiled == (
        b"SCgf\x00\x00\x00\x02\x00\x01\x13system_link_audio_2\x00\x00\x00\x05\x00\x00"
        b"\x00\x00?\x80\x00\x00@\x00\x00\x00\xc2\xc6\x00\x00@@\x00\x00\x00\x00"
        b"\x00\x05@\x00\x00\x00<\xa3\xd7\n?\x80\x00\x00A\x80\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x05\x0bdone_action\x00\x00\x00\x00\tfade_time"
        b"\x00\x00\x00\x01\x04gate\x00\x00\x00\x02\x03in_\x00\x00\x00\x03\x03ou"
        b"t\x00\x00\x00\x04\x00\x00\x00\x07\x07Control\x01\x00\x00\x00\x00\x00\x00"
        b"\x00\x05\x00\x00\x01\x01\x01\x01\x01\x0cBinaryOpUGen\x01\x00\x00\x00\x02\x00"
        b"\x00\x00\x01\x00\n\x00\x00\x00\x00\x00\x00\x00\x01\xff\xff\xff"
        b"\xff\x00\x00\x00\x00\x01\x06EnvGen\x01\x00\x00\x00\x11\x00\x00"
        b"\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\xff\xff\xff\xff"
        b"\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        b"\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x02\xff\xff\xff\xff"
        b"\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x03\xff\xff\xff\xff"
        b"\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x01\xff\xff\xff\xff"
        b"\x00\x00\x00\x04\xff\xff\xff\xff\x00\x00\x00\x00\xff\xff\xff\xff"
        b"\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x01\xff\xff\xff\xff"
        b"\x00\x00\x00\x04\xff\xff\xff\xff\x00\x00\x00\x00\x01\nInFeedback"
        b"\x02\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x03\x02\x02\x0cBinaryOpUGen\x02\x00\x00\x00\x02\x00\x00\x00\x01\x00"
        b"\x02\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00"
        b"\x00\x02\x0cBinaryOpUGen\x02\x00\x00\x00\x02\x00\x00\x00\x01\x00\x02\x00\x00"
        b"\x00\x03\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x00\x02\x03Out\x02"
        b"\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x05\x00\x00"
        b"\x00\x00\x00\x00"
    )


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_system_link_audio_2_sclang():
    sc_compiled = SuperColliderSynthDef(
        "system_link_audio_2",
        r"""
        | out=0, in=16, vol=1, level=1, lag=0.05, doneAction=2 |
        var env = EnvGate(i_level: 0, doneAction:doneAction, curve:'sin')
        * Lag.kr(vol * level, lag);
        Out.ar(out, InFeedback.ar(in, 2) * env)
        """,
        r"[\kr, \kr, \kr, \kr, \kr, \ir]",
    ).compile()
    sc_synthdef = decompile_synthdef(sc_compiled)
    assert normalize(str(sc_synthdef)) == normalize(
        """
        synthdef:
            name: system_link_audio_2
            ugens:
            -   Control.ir:
                    doneAction: 2.0
            -   Control.kr:
                    out: 0.0
                    in: 16.0
                    vol: 1.0
                    level: 1.0
                    lag: 0.05000000074505806
            -   BinaryOpUGen(MULTIPLICATION).kr/0:
                    left: Control.kr[2:vol]
                    right: Control.kr[3:level]
            -   Lag.kr:
                    source: BinaryOpUGen(MULTIPLICATION).kr/0[0]
                    lag_time: Control.kr[4:lag]
            -   InFeedback.ar:
                    bus: Control.kr[1:in]
            -   Control.kr:
                    gate: 1.0
            -   Control.kr:
                    fadeTime: 0.019999999552965164
            -   EnvGen.kr:
                    gate: Control.kr[0:gate]
                    level_scale: 1.0
                    level_bias: 0.0
                    time_scale: Control.kr[0:fadeTime]
                    done_action: Control.ir[0:doneAction]
                    envelope[0]: 0.0
                    envelope[1]: 2.0
                    envelope[2]: 1.0
                    envelope[3]: -99.0
                    envelope[4]: 1.0
                    envelope[5]: 1.0
                    envelope[6]: 3.0
                    envelope[7]: 0.0
                    envelope[8]: 0.0
                    envelope[9]: 1.0
                    envelope[10]: 3.0
                    envelope[11]: 0.0
            -   BinaryOpUGen(MULTIPLICATION).kr/1:
                    left: EnvGen.kr[0]
                    right: Lag.kr[0]
            -   BinaryOpUGen(MULTIPLICATION).ar/0:
                    left: InFeedback.ar[0]
                    right: BinaryOpUGen(MULTIPLICATION).kr/1[0]
            -   BinaryOpUGen(MULTIPLICATION).ar/1:
                    left: InFeedback.ar[1]
                    right: BinaryOpUGen(MULTIPLICATION).kr/1[0]
            -   Out.ar:
                    bus: Control.kr[0:out]
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                    source[1]: BinaryOpUGen(MULTIPLICATION).ar/1[0]
        """
    )



from typing import Dict, Generator, Iterator, Optional, Sequence, Tuple, Union
from uuid import UUID

from uqbar.enums import IntEnumeration

from .patterns import Pattern, SequencePattern, T


class ChoicePattern(SequencePattern[T]):
    def __init__(
        self,
        sequence: Sequence[Union[T, Pattern[T]]],
        iterations: Optional[int] = 1,
        forbid_repetitions: bool = False,
        weights: Optional[Sequence[float]] = None,
    ) -> None:
        super().__init__(sequence, iterations=iterations)
        self._forbid_repetitions = bool(forbid_repetitions)
        self._weights = tuple(abs(float(x)) for x in weights) if weights else None

    def _iterate(
        self, state: Optional[Dict[str, UUID]] = None
    ) -> Generator[T, bool, None]:
        should_stop = False
        rng = self._get_rng()
        previous_index: Optional[int] = None
        for _ in self._loop(self._iterations):
            if self._weights:
                index = self._find_index_weighted(rng, self._weights)
                while self.forbid_repetitions and index == previous_index:
                    index = self._find_index_weighted(rng, self._weights)
            else:
                index = self._find_index_unweighted(rng)
                while self.forbid_repetitions and index == previous_index:
                    index = self._find_index_unweighted(rng)
            previous_index = index
            choice = self._sequence[index]
            if isinstance(choice, Pattern):
                # MyPy: Function does not return a value (it only ever returns None)
                should_stop = (yield from choice) or should_stop  # type: ignore
            else:
                should_stop = (yield choice) or should_stop
            if should_stop:
                break
        return

    def _find_index_unweighted(self, rng: Iterator[float]) -> int:
        return int(next(rng) * 0x7FFFFFFF) % len(self._sequence)

    def _find_index_weighted(
        self, rng: Iterator[float], weights: Sequence[float]
    ) -> int:
        needle = next(rng) * sum(weights)
        sum_of_weights = 0.0
        for index, weight in enumerate(weights):
            if sum_of_weights <= needle <= sum_of_weights + weight:
                break
            sum_of_weights += weight
        return index

    @property
    def forbid_repetitions(self) -> bool:
        return self._forbid_repetitions

    @property
    def weights(self) -> Optional[Tuple[float, ...]]:
        return self._weights


class RandomPattern(Pattern[float]):
    class Distribution(IntEnumeration):
        WHITE_NOISE = 0

    def __init__(
        self,
        minimum: float = 0.0,
        maximum: float = 1.0,
        iterations: Optional[int] = None,
        distribution: Union["RandomPattern.Distribution", str] = "WHITE_NOISE",
    ) -> None:
        if iterations is not None:
            iterations = int(iterations)
            if iterations < 1:
                raise ValueError("Iterations must be null or greater than 0")
        self._iterations = iterations
        self._distribution: RandomPattern.Distribution = self.Distribution.from_expr(
            distribution
        )
        self._minimum: float = self._freeze_recursive(minimum)
        self._maximum: float = self._freeze_recursive(maximum)

    def _iterate(
        self, state: Optional[Dict[str, UUID]] = None
    ) -> Generator[float, bool, None]:
        def procedure(one: float, two: float) -> float:
            minimum, maximum = sorted([one, two])
            number = next(rng)
            return (number * (maximum - minimum)) + minimum

        rng = self._get_rng()
        for _ in self._loop(self._iterations):
            expr = self._apply_recursive(procedure, self._minimum, self._maximum)
            if (yield expr):
                return

    @property
    def distribution(self) -> Union["RandomPattern.Distribution", str]:
        return self._distribution

    @property
    def is_infinite(self) -> bool:
        return self._iterations is None

    @property
    def iterations(self) -> Optional[int]:
        return self._iterations

    @property
    def minimum(self) -> float:
        return self._minimum

    @property
    def maximum(self) -> float:
        return self._maximum


class ShufflePattern(SequencePattern[T]):
    def __init__(
        self,
        sequence: Sequence[Union[T, Pattern[T]]],
        iterations: Optional[int] = 1,
        forbid_repetitions: bool = False,
    ) -> None:
        super().__init__(sequence, iterations=iterations)
        self._forbid_repetitions = bool(forbid_repetitions)

    def _iterate(
        self, state: Optional[Dict[str, UUID]] = None
    ) -> Generator[T, bool, None]:
        should_stop = False
        rng = self._get_rng()
        previous_index = None
        for _ in self._loop(self._iterations):
            indices = self._shuffle(len(self._sequence), rng)
            while (
                self.forbid_repetitions
                and len(indices) > 1
                and indices[0] == previous_index
            ):
                indices = self._shuffle(len(self._sequence), rng)
            previous_index = indices[-1]
            for index in indices:
                choice = self._sequence[index]
                if isinstance(choice, Pattern):
                    # MyPy: Function does not return a value (it only ever returns None)
                    should_stop = (yield from choice) or should_stop  # type: ignore
                else:
                    should_stop = (yield choice) or should_stop
                if should_stop:
                    return

    def _shuffle(self, length: int, rng: Iterator[float]) -> Sequence[int]:
        indices = list(range(length))
        shuffled_indices = []
        while len(indices) > 1:
            index = int(next(rng) * 0x7FFFFFFF) % len(indices)
            shuffled_indices.append(indices.pop(index))
        if indices:
            shuffled_indices.append(indices.pop())
        return shuffled_indices

    @property
    def forbid_repetitions(self) -> bool:
        return self._forbid_repetitions



import pytest

from supriya import AddAction, CalculationRate
from supriya.assets import synthdefs
from supriya.patterns import (
    BusAllocateEvent,
    BusFreeEvent,
    BusPattern,
    CompositeEvent,
    EventPattern,
    GroupAllocateEvent,
    NodeFreeEvent,
    NoteEvent,
    NullEvent,
    SequencePattern,
    SynthAllocateEvent,
)
from supriya.patterns.testutils import MockUUID as M
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, inner_pattern, calculation_rate, channel_count, release_time, expected, is_infinite",
    [
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            "audio",
            2,
            0.0,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            M("A"),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(M("B")),
                        SynthAllocateEvent(
                            M("C"),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.0,
                            in_=M("A"),
                            synthdef=synthdefs.system_link_audio_2,
                            target_node=M("B"),
                        ),
                    ]
                ),
                NoteEvent(M("D"), a=1, out=M("A"), target_node=M("B")),
                NoteEvent(M("E"), a=2, out=M("A"), target_node=M("B")),
                CompositeEvent(
                    [NodeFreeEvent(M("C")), NodeFreeEvent(M("B")), BusFreeEvent(M("A"))]
                ),
            ],
            False,
        ),
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            "audio",
            2,
            0.25,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            M("A"),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(M("B")),
                        SynthAllocateEvent(
                            M("C"),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=M("A"),
                            synthdef=synthdefs.system_link_audio_2,
                            target_node=M("B"),
                        ),
                    ]
                ),
                NoteEvent(M("D"), a=1, out=M("A"), target_node=M("B")),
                NoteEvent(M("E"), a=2, out=M("A"), target_node=M("B")),
                CompositeEvent(
                    [
                        NodeFreeEvent(M("C")),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(M("B")),
                        BusFreeEvent(M("A")),
                    ]
                ),
            ],
            False,
        ),
        (
            None,
            BusPattern(EventPattern(a=SequencePattern([1, 2])), channel_count=2),
            "audio",
            2,
            0.25,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            M("A"),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(M("B")),
                        SynthAllocateEvent(
                            M("C"),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=M("A"),
                            synthdef=synthdefs.system_link_audio_2,
                            target_node=M("B"),
                        ),
                    ]
                ),
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            M("D"),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(M("E"), target_node=M("B")),
                        SynthAllocateEvent(
                            M("F"),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=M("D"),
                            out=M("A"),
                            synthdef=synthdefs.system_link_audio_2,
                            target_node=M("E"),
                        ),
                    ]
                ),
                NoteEvent(M("G"), a=1, out=M("D"), target_node=M("E")),
                NoteEvent(M("H"), a=2, out=M("D"), target_node=M("E")),
                CompositeEvent(
                    [
                        NodeFreeEvent(M("F")),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(M("E")),
                        BusFreeEvent(M("D")),
                    ]
                ),
                CompositeEvent(
                    [
                        NodeFreeEvent(M("C")),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(M("B")),
                        BusFreeEvent(M("A")),
                    ]
                ),
            ],
            False,
        ),
    ],
)
def test(
    stop_at,
    inner_pattern,
    calculation_rate,
    channel_count,
    release_time,
    expected,
    is_infinite,
):
    pattern = BusPattern(
        inner_pattern,
        calculation_rate=calculation_rate,
        channel_count=channel_count,
        release_time=release_time,
    )
    run_pattern_test(pattern, expected, is_infinite, stop_at)



from supriya.ugens import (
    Decay2,
    DelayC,
    In,
    Mix,
    Out,
    Parameter,
    SinOsc,
    SynthDefBuilder,
    decompile_synthdef,
)


def test_SynthDefDecompiler_01():
    r"""Anonymous SynthDef without parameters."""
    with SynthDefBuilder() as builder:
        sine = SinOsc.ar()
        sine = -sine
        Out.ar(bus=99, source=sine)
    old_synthdef = builder.build()
    compiled_synthdef = old_synthdef.compile()
    new_synthdef = decompile_synthdef(compiled_synthdef)
    assert str(old_synthdef) == str(new_synthdef)
    assert old_synthdef.indexed_parameters == new_synthdef.indexed_parameters
    assert compiled_synthdef == new_synthdef.compile()
    assert old_synthdef.anonymous_name == new_synthdef.anonymous_name
    assert old_synthdef.name == new_synthdef.name


def test_SynthDefDecompiler_02():
    r"""Anonymous SynthDef with one parameter."""
    with SynthDefBuilder(freq=440) as builder:
        sine = SinOsc.ar(frequency=builder["freq"])
        Out.ar(bus=0, source=sine)
    old_synthdef = builder.build()
    compiled_synthdef = old_synthdef.compile()
    new_synthdef = decompile_synthdef(compiled_synthdef)
    assert str(old_synthdef) == str(new_synthdef)
    assert old_synthdef.indexed_parameters == new_synthdef.indexed_parameters
    assert compiled_synthdef == new_synthdef.compile()
    assert old_synthdef.anonymous_name == new_synthdef.anonymous_name
    assert old_synthdef.name == new_synthdef.name


def test_SynthDefDecompiler_03():
    r"""Named SynthDef with one parameter."""
    with SynthDefBuilder(freq=440) as builder:
        sine = SinOsc.ar(frequency=builder["freq"])
        Out.ar(bus=0, source=sine)
    old_synthdef = builder.build("test")
    compiled_synthdef = old_synthdef.compile()
    new_synthdef = decompile_synthdef(compiled_synthdef)
    assert str(old_synthdef) == str(new_synthdef)
    assert old_synthdef.indexed_parameters == new_synthdef.indexed_parameters
    assert compiled_synthdef == new_synthdef.compile()
    assert old_synthdef.anonymous_name == new_synthdef.anonymous_name
    assert old_synthdef.name == new_synthdef.name


def test_SynthDefDecompiler_04():
    r"""Multiple parameters."""
    with SynthDefBuilder(freq=1200, out=23) as builder:
        sine = SinOsc.ar(frequency=builder["freq"])
        Out.ar(bus=builder["out"], source=sine)
    old_synthdef = builder.build("test")
    compiled_synthdef = old_synthdef.compile()
    new_synthdef = decompile_synthdef(compiled_synthdef)
    assert str(old_synthdef) == str(new_synthdef)
    assert old_synthdef.indexed_parameters == new_synthdef.indexed_parameters
    assert compiled_synthdef == new_synthdef.compile()
    assert old_synthdef.anonymous_name == new_synthdef.anonymous_name
    assert old_synthdef.name == new_synthdef.name


def test_SynthDefDecompiler_05():
    """Multiple parameters."""
    builder = SynthDefBuilder(damping=0.5, delay_time=1.0, room_size=0.75)
    with builder:
        microphone = In.ar(bus=0)
        delay = DelayC.ar(
            source=microphone, maximum_delay_time=5.0, delay_time=builder["delay_time"]
        )
        Out.ar(bus=0, source=delay)
    old_synthdef = builder.build("test")
    compiled_synthdef = old_synthdef.compile()
    new_synthdef = decompile_synthdef(compiled_synthdef)
    assert str(old_synthdef) == str(new_synthdef)
    assert old_synthdef.indexed_parameters == new_synthdef.indexed_parameters
    assert compiled_synthdef == new_synthdef.compile()
    assert old_synthdef.anonymous_name == new_synthdef.anonymous_name
    assert old_synthdef.name == new_synthdef.name


def test_SynthDefDecompiler_06():
    r"""Multiple parameters with different rates."""
    builder = SynthDefBuilder(
        a_phase=0.0, freq=440, i_decay_time=1.0, t_trig_a=0, t_trig_b=0
    )
    with builder:
        decay = Decay2.kr(
            source=(builder["t_trig_a"], builder["t_trig_b"]),
            attack_time=0.5,
            decay_time=builder["i_decay_time"],
        )
        sin_osc = SinOsc.ar(frequency=builder["freq"], phase=builder["a_phase"])
        enveloped_sin_osc = sin_osc * decay
        Out.ar(bus=0, source=enveloped_sin_osc)
    old_synthdef = builder.build("trigTest")
    compiled_synthdef = old_synthdef.compile()
    new_synthdef = decompile_synthdef(compiled_synthdef)
    assert str(old_synthdef) == str(new_synthdef)
    assert old_synthdef.indexed_parameters == new_synthdef.indexed_parameters
    assert compiled_synthdef == new_synthdef.compile()
    assert old_synthdef.anonymous_name == new_synthdef.anonymous_name
    assert old_synthdef.name == new_synthdef.name


def test_SynthDefDecompiler_07():
    builder = SynthDefBuilder(amp=0.5, freqs=[300, 400])
    with builder:
        sines = SinOsc.ar(frequency=builder["freqs"])
        sines = Mix.new(sines)
        sines = sines * builder["amp"]
        Out.ar(bus=0, source=sines)
    old_synthdef = builder.build("arrayarg")
    compiled_synthdef = old_synthdef.compile()
    new_synthdef = decompile_synthdef(compiled_synthdef)
    assert str(old_synthdef) == str(new_synthdef)
    assert old_synthdef.indexed_parameters == new_synthdef.indexed_parameters
    assert compiled_synthdef == new_synthdef.compile()
    assert old_synthdef.anonymous_name == new_synthdef.anonymous_name
    assert old_synthdef.name == new_synthdef.name


def test_SynthDefDecompiler_08():
    r"""Multiple parameters with different lags."""
    builder = SynthDefBuilder(amp=0.5, freqs=Parameter(lag=0.5, value=[300, 400]))
    with builder:
        sines = SinOsc.ar(frequency=builder["freqs"])
        sines = Mix.new(sines)
        sines = sines * builder["amp"]
        Out.ar(bus=0, source=sines)
    old_synthdef = builder.build("arrayarg")
    compiled_synthdef = old_synthdef.compile()
    new_synthdef = decompile_synthdef(compiled_synthdef)
    assert str(old_synthdef) == str(new_synthdef)
    assert old_synthdef.indexed_parameters == new_synthdef.indexed_parameters
    assert compiled_synthdef == new_synthdef.compile()
    assert old_synthdef.anonymous_name == new_synthdef.anonymous_name
    assert old_synthdef.name == new_synthdef.name



from collections.abc import Sequence
from typing import Any, Dict, Tuple

from ..enums import CalculationRate
from .core import UGen, param, ugen


@ugen(dr=True)
class Dbrown(UGen):
    """
    A demand-rate brownian movement generator.

    ::

        >>> dbrown = supriya.ugens.Dbrown.dr(
        ...     length=float("inf"),
        ...     maximum=1,
        ...     minimum=0,
        ...     step=0.01,
        ... )
        >>> dbrown
        <Dbrown.dr()[0]>
    """

    minimum = param(0.0)
    maximum = param(1.0)
    step = param(0.01)
    length = param(float("inf"))


@ugen(dr=True)
class Dbufrd(UGen):
    """
    A buffer-reading demand-rate UGen.

    ::

        >>> dbufrd = supriya.ugens.Dbufrd.dr(
        ...     buffer_id=0,
        ...     loop=1,
        ...     phase=0,
        ... )
        >>> dbufrd
        <Dbufrd.dr()[0]>
    """

    buffer_id = param(0)
    phase = param(0)
    loop = param(1)


@ugen(dr=True)
class Dbufwr(UGen):
    """
    A buffer-writing demand-rate UGen.

    ::

        >>> dbufwr = supriya.ugens.Dbufwr.dr(
        ...     buffer_id=0,
        ...     source=0,
        ...     loop=1,
        ...     phase=0,
        ... )
        >>> dbufwr
        <Dbufwr.dr()[0]>
    """

    source = param(0.0)
    buffer_id = param(0.0)
    phase = param(0.0)
    loop = param(1.0)


@ugen(ar=True, kr=True)
class Demand(UGen):
    """
    Demands results from demand-rate UGens.

    ::

        >>> source = [
        ...     supriya.ugens.Dseries.dr(),
        ...     supriya.ugens.Dwhite.dr(),
        ... ]
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1)
        >>> demand = supriya.ugens.Demand.ar(
        ...     reset=0,
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> demand
        <Demand.ar()>
    """

    trigger = param(0)
    reset = param(0)
    source = param(unexpanded=True)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs,
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        if not isinstance(source := kwargs["source"], Sequence):
            kwargs["source"] = [source]
        self._channel_count = len(kwargs["source"])
        return calculation_rate, kwargs


@ugen(ar=True, kr=True)
class DemandEnvGen(UGen):
    """
    A demand rate envelope generator.

    ::

        >>> demand_env_gen = supriya.ugens.DemandEnvGen.ar(
        ...     curve=0,
        ...     done_action=0,
        ...     duration=1,
        ...     gate=1,
        ...     level=1,
        ...     level_bias=0,
        ...     level_scale=1,
        ...     reset=1,
        ...     shape=1,
        ...     time_scale=1,
        ... )
        >>> demand_env_gen
        <DemandEnvGen.ar()[0]>
    """

    level = param()
    duration = param()
    shape = param(1)
    curve = param(0)
    gate = param(1)
    reset = param(1)
    level_scale = param(1)
    level_bias = param(0)
    time_scale = param(1)
    done_action = param(0)


@ugen(dr=True)
class Dgeom(UGen):
    """
    A demand-rate geometric series generator.

    ::

        >>> dgeom = supriya.ugens.Dgeom.dr(
        ...     grow=2,
        ...     length=float("inf"),
        ...     start=1,
        ... )
        >>> dgeom
        <Dgeom.dr()[0]>
    """

    start = param(1)
    grow = param(2)
    length = param(float("inf"))


@ugen(dr=True)
class Dibrown(UGen):
    """
    An integer demand-rate brownian movement generator.

    ::

        >>> dibrown = supriya.ugens.Dibrown.dr(
        ...     length=float("inf"),
        ...     maximum=1,
        ...     minimum=0,
        ...     step=0.01,
        ... )
        >>> dibrown
        <Dibrown.dr()[0]>
    """

    minimum = param(0)
    maximum = param(12)
    step = param(1)
    length = param(float("inf"))


@ugen(dr=True)
class Diwhite(UGen):
    """
    An integer demand-rate white noise random generator.

    ::

        >>> diwhite = supriya.ugens.Diwhite.dr(
        ...     length=float("inf"),
        ...     maximum=1,
        ...     minimum=0,
        ... )
        >>> diwhite
        <Diwhite.dr()[0]>
    """

    minimum = param(0)
    maximum = param(1)
    length = param(float("inf"))


@ugen(dr=True)
class Drand(UGen):
    """
    A demand-rate random sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> drand = supriya.ugens.Drand.dr(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> drand
        <Drand.dr()[0]>
    """

    repeats = param(1)
    sequence = param(unexpanded=True)


@ugen(dr=True)
class Dreset(UGen):
    """
    Resets demand-rate UGens.

    ::

        >>> source = supriya.ugens.Dseries.dr(start=0, step=2)
        >>> dreset = supriya.ugens.Dreset.dr(
        ...     reset=0,
        ...     source=source,
        ... )
        >>> dreset
        <Dreset.dr()[0]>
    """

    source = param()
    reset = param(0)


@ugen(dr=True)
class Dseq(UGen):
    """
    A demand-rate sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> dseq = supriya.ugens.Dseq.dr(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> dseq
        <Dseq.dr()[0]>
    """

    repeats = param(1)
    sequence = param(unexpanded=True)


@ugen(dr=True)
class Dser(UGen):
    """
    A demand-rate sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> dser = supriya.ugens.Dser.dr(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> dser
        <Dser.dr()[0]>
    """

    repeats = param(1)
    sequence = param(unexpanded=True)


@ugen(dr=True)
class Dseries(UGen):
    """
    A demand-rate arithmetic series.

    ::

        >>> dseries = supriya.ugens.Dseries.dr(
        ...     length=float("inf"),
        ...     start=1,
        ...     step=1,
        ... )
        >>> dseries
        <Dseries.dr()[0]>
    """

    length = param(float("inf"))
    start = param(1)
    step = param(1)


@ugen(dr=True)
class Dshuf(UGen):
    """
    A demand-rate random sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> dshuf = supriya.ugens.Dshuf.dr(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> dshuf
        <Dshuf.dr()[0]>
    """

    repeats = param(1)
    sequence = param(unexpanded=True)


@ugen(dr=True)
class Dstutter(UGen):
    """
    A demand-rate input replicator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> dstutter = supriya.ugens.Dstutter.dr(
        ...     n=2,
        ...     source=source,
        ... )
        >>> dstutter
        <Dstutter.dr()[0]>
    """

    n = param(2)
    source = param()


@ugen(dr=True)
class Dswitch(UGen):
    """
    A demand-rate generator for embedding different inputs.

    ::

        >>> index = supriya.ugens.Dseq.dr(sequence=[0, 1, 2, 1, 0])
        >>> sequence = (1.0, 2.0, 3.0)
        >>> dswitch = supriya.ugens.Dswitch.dr(
        ...     index_=index,
        ...     sequence=sequence,
        ... )
        >>> dswitch
        <Dswitch.dr()[0]>
    """

    index_ = param()
    sequence = param(unexpanded=True)


@ugen(dr=True)
class Dswitch1(UGen):
    """
    A demand-rate generator for switching between inputs.

    ::

        >>> index = supriya.ugens.Dseq.dr(sequence=[0, 1, 2, 1, 0])
        >>> sequence = (1.0, 2.0, 3.0)
        >>> dswitch_1 = supriya.ugens.Dswitch1.dr(
        ...     index_=index,
        ...     sequence=sequence,
        ... )
        >>> dswitch_1
        <Dswitch1.dr()[0]>
    """

    index_ = param()
    sequence = param(unexpanded=True)


@ugen(dr=True)
class Dunique(UGen):
    """
    Returns the same unique series of values for several demand streams.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> dunique = supriya.ugens.Dunique.dr(
        ...     max_buffer_size=1024,
        ...     protected=True,
        ...     source=source,
        ... )
        >>> dunique
        <Dunique.dr()[0]>
    """

    source = param()
    max_buffer_size = param(1024)
    protected = param(True)


@ugen(ar=True, kr=True)
class Duty(UGen):
    """
    A value is demanded of each UGen in the list and output according to a stream of duration values.

    ::

        >>> duty = supriya.ugens.Duty.kr(
        ...     done_action=0,
        ...     duration=supriya.ugens.Drand.dr(
        ...         sequence=[0.01, 0.2, 0.4],
        ...         repeats=2,
        ...     ),
        ...     reset=0,
        ...     level=supriya.ugens.Dseq.dr(
        ...         sequence=[204, 400, 201, 502, 300, 200],
        ...         repeats=2,
        ...     ),
        ... )
        >>> duty
        <Duty.kr()[0]>
    """

    duration = param(1.0)
    reset = param(0.0)
    level = param(1.0)
    done_action = param(0.0)


@ugen(dr=True)
class Dwhite(UGen):
    """
    A demand-rate white noise random generator.

    ::

        >>> dwhite = supriya.ugens.Dwhite.dr(
        ...     length=float("inf"),
        ...     maximum=1,
        ...     minimum=0,
        ... )
        >>> dwhite
        <Dwhite.dr()[0]>
    """

    minimum = param(0.0)
    maximum = param(0.0)
    length = param(float("inf"))


@ugen(dr=True)
class Dwrand(UGen):
    """
    A demand-rate weighted random sequence generator.

    ::

        >>> sequence = [0, 1, 2, 7]
        >>> weights = [0.4, 0.4, 0.1, 0.1]
        >>> dwrand = supriya.ugens.Dwrand.dr(
        ...     repeats=1,
        ...     sequence=sequence,
        ...     weights=weights,
        ... )
        >>> dwrand
        <Dwrand.dr()[0]>
    """

    repeats = param(1)
    length = param()
    weights = param(unexpanded=True)
    sequence = param(unexpanded=True)

    @classmethod
    def dr(cls, repeats=1, sequence=None, weights=None):
        if not isinstance(sequence, Sequence):
            sequence = [sequence]
        sequence = tuple(float(_) for _ in sequence)
        if not isinstance(weights, Sequence):
            weights = [weights]
        weights = tuple(float(_) for _ in weights)
        weights = weights[: len(sequence)]
        weights += (0.0,) * (len(sequence) - len(weights))
        return cls._new_expanded(
            calculation_rate=CalculationRate.DEMAND,
            repeats=repeats,
            length=len(sequence),
            sequence=sequence,
            weights=weights,
        )


@ugen(dr=True)
class Dxrand(UGen):
    """
    A demand-rate random sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> dxrand = supriya.ugens.Dxrand.dr(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> dxrand
        <Dxrand.dr()[0]>
    """

    repeats = param(1)
    sequence = param(unexpanded=True)



# flake8: noqa
import os
import platform

import pytest

from supriya.ugens import (
    DetectSilence,
    FreeSelf,
    In,
    Out,
    SinOsc,
    SuperColliderSynthDef,
    SynthDefBuilder,
)


@pytest.fixture
def py_synthdef_01():
    with SynthDefBuilder() as builder:
        sine_one = SinOsc.ar(frequency=420)
        sine_two = SinOsc.ar(frequency=440)
        sines = sine_one * sine_two
        Out.ar(bus=0, source=sines)
    py_synthdef = builder.build("foo")
    return py_synthdef


def test_basic_01_supriya_vs_bytes(py_synthdef_01):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x03foo'
                b'\x00\x00\x00\x03'
                    b'C\xd2\x00\x00'
                    b'\x00\x00\x00\x00'
                    b'C\xdc\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x04'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x02'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x02'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x02'
                    b'\x0cBinaryOpUGen'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x02'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x02'
                            b'\x00\x00\x00\x00'
                b'\x00\x00',
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_01.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_basic_01_supriya_vs_sclang(py_synthdef_01):
    sc_synthdef = SuperColliderSynthDef(
        "foo", "Out.ar(0, SinOsc.ar(freq: 420) * SinOsc.ar(freq: 440))"
    )
    sc_compiled_synthdef = sc_synthdef.compile()
    py_compiled_synthdef = py_synthdef_01.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


@pytest.fixture
def py_synthdef_02():
    with SynthDefBuilder() as builder:
        sine = SinOsc.ar()
        sine = -sine
        Out.ar(bus=99, source=sine)
    py_synthdef = builder.build("test")
    return py_synthdef


def test_basic_02_supriya_vs_bytes(py_synthdef_02):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x04test'
                b'\x00\x00\x00\x03'
                    b'C\xdc\x00\x00'
                    b'\x00\x00\x00\x00'
                    b'B\xc6\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x03'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x02'
                    b'\x0bUnaryOpUGen'
                        b'\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x02'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                b'\x00\x00',
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_02.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_basic_02_supriya_vs_sclang(py_synthdef_02):
    sc_synthdef = SuperColliderSynthDef("test", "Out.ar(99, SinOsc.ar(freq: 440).neg)")
    sc_compiled_synthdef = sc_synthdef.compile()
    py_compiled_synthdef = py_synthdef_02.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


@pytest.fixture
def py_synthdef_03():
    with SynthDefBuilder() as builder:
        inputs = In.ar(bus=8, channel_count=2)
        Out.ar(bus=0, source=inputs)
    py_synthdef = builder.build("test")
    return py_synthdef


def test_basic_03_supriya_vs_bytes(py_synthdef_03):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x04test'
                b'\x00\x00\x00\x02'
                    b'A\x00\x00\x00'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x02'
                    b'\x02In'
                        b'\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\x02'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x01'
                b'\x00\x00',
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_03.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_basic_03_supriya_vs_sclang(py_synthdef_03):
    sc_synthdef = SuperColliderSynthDef(
        "test",
        r"""
        Out.ar(0, In.ar(8, 2))
        """,
    )
    sc_compiled_synthdef = sc_synthdef.compile()
    py_compiled_synthdef = py_synthdef_03.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


@pytest.fixture
def py_synthdef_04():
    with SynthDefBuilder() as builder:
        sin_osc = SinOsc.ar()
        FreeSelf.kr(trigger=sin_osc)
        Out.ar(bus=0, source=sin_osc)
    py_synthdef = builder.build("test")
    return py_synthdef


def test_basic_04_supriya_vs_bytes(py_synthdef_04):
    """
    FreeSelf.
    """
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x04test'
                b'\x00\x00\x00\x02'
                    b'C\xdc\x00\x00'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x03'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x02'
                    b'\x08FreeSelf'
                        b'\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x01'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                b'\x00\x00',
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_04.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_basic_04_supriya_vs_sclang(py_synthdef_04):
    sc_synthdef = SuperColliderSynthDef(
        "test",
        r"""
        Out.ar(0, FreeSelf.kr(SinOsc.ar()))
        """,
    )
    sc_compiled_synthdef = bytes(sc_synthdef.compile())
    py_compiled_synthdef = py_synthdef_04.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_basic_05_supriya_vs_sclang():
    with SynthDefBuilder() as builder:
        source = In.ar(bus=8, channel_count=2)
        DetectSilence.ar(source=source)
        Out.ar(bus=0, source=source)
    py_synthdef = builder.build("DetectSilenceTest")
    py_compiled_synthdef = py_synthdef.compile()
    sc_synthdef = SuperColliderSynthDef(
        "DetectSilenceTest",
        r"""
        var source, detect_silence, out;
        source = In.ar(8, 2);
        detect_silence = DetectSilence.ar(source);
        out = Out.ar(0, source);
        """,
    )
    sc_compiled_synthdef = bytes(sc_synthdef.compile())
    assert py_compiled_synthdef == sc_compiled_synthdef



from uqbar.strings import normalize

from supriya.ugens import SinOsc
from supriya.ugens.factories import SynthDefFactory


def test_gate_01():
    def signal_block(builder, source, state):
        return SinOsc.ar()

    factory = SynthDefFactory(channel_count=1)
    factory = factory.with_signal_block(signal_block)
    factory = factory.with_gate()
    factory = factory.with_output()
    assert str(factory.build(name="test")) == normalize(
        """
        synthdef:
            name: test
            ugens:
            -   Control.kr:
                    gate: 1.0
            -   Linen.kr:
                    gate: Control.kr[0:gate]
                    attack_time: 0.02
                    sustain_level: 1.0
                    release_time: 0.02
                    done_action: 2.0
            -   Control.ir:
                    out: 0.0
            -   SinOsc.ar:
                    frequency: 440.0
                    phase: 0.0
            -   BinaryOpUGen(MULTIPLICATION).ar:
                    left: SinOsc.ar[0]
                    right: Linen.kr[0]
            -   Out.ar:
                    bus: Control.ir[0:out]
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
        """
    )


def test_gate_02():
    def signal_block(builder, source, state):
        return SinOsc.ar()

    factory = SynthDefFactory(channel_count=1)
    factory = factory.with_signal_block(signal_block)
    factory = factory.with_gate()
    factory = factory.with_output(crossfaded=True)
    assert str(factory.build(name="test")) == normalize(
        """
        synthdef:
            name: test
            ugens:
            -   Control.kr:
                    gate: 1.0
                    mix: 0.0
            -   Linen.kr:
                    gate: Control.kr[0:gate]
                    attack_time: 0.02
                    sustain_level: 1.0
                    release_time: 0.02
                    done_action: 2.0
            -   BinaryOpUGen(MULTIPLICATION).kr:
                    left: Control.kr[1:mix]
                    right: Linen.kr[0]
            -   Control.ir:
                    out: 0.0
            -   SinOsc.ar:
                    frequency: 440.0
                    phase: 0.0
            -   XOut.ar:
                    bus: Control.ir[0:out]
                    crossfade: BinaryOpUGen(MULTIPLICATION).kr[0]
                    source[0]: SinOsc.ar[0]
        """
    )


def test_gate_03():
    def signal_block(builder, source, state):
        return SinOsc.ar()

    factory = SynthDefFactory(channel_count=1)
    factory = factory.with_signal_block(signal_block)
    factory = factory.with_gate()
    factory = factory.with_output(crossfaded=True, windowed=True)
    assert str(factory.build(name="test")) == normalize(
        """
        synthdef:
            name: test
            ugens:
            -   Control.kr:
                    gate: 1.0
            -   Linen.kr:
                    gate: Control.kr[0:gate]
                    attack_time: 0.02
                    sustain_level: 1.0
                    release_time: 0.02
                    done_action: 2.0
            -   Control.ir:
                    duration: 1.0
                    out: 0.0
            -   Line.kr:
                    start: 0.0
                    stop: 1.0
                    duration: Control.ir[0:duration]
                    done_action: 2.0
            -   UnaryOpUGen(HANNING_WINDOW).kr:
                    source: Line.kr[0]
            -   BinaryOpUGen(MULTIPLICATION).kr:
                    left: UnaryOpUGen(HANNING_WINDOW).kr[0]
                    right: Linen.kr[0]
            -   SinOsc.ar:
                    frequency: 440.0
                    phase: 0.0
            -   XOut.ar:
                    bus: Control.ir[1:out]
                    crossfade: BinaryOpUGen(MULTIPLICATION).kr[0]
                    source[0]: SinOsc.ar[0]
        """
    )


def test_gate_04():
    def signal_block(builder, source, state):
        return SinOsc.ar()

    factory = SynthDefFactory(channel_count=1)
    factory = factory.with_signal_block(signal_block)
    factory = factory.with_gate()
    factory = factory.with_output(crossfaded=True, leveled=True, windowed=True)
    assert str(factory.build(name="test")) == normalize(
        """
        synthdef:
            name: test
            ugens:
            -   Control.kr:
                    gate: 1.0
                    level: 1.0
            -   Linen.kr:
                    gate: Control.kr[0:gate]
                    attack_time: 0.02
                    sustain_level: 1.0
                    release_time: 0.02
                    done_action: 2.0
            -   Control.ir:
                    duration: 1.0
                    out: 0.0
            -   Line.kr:
                    start: 0.0
                    stop: 1.0
                    duration: Control.ir[0:duration]
                    done_action: 2.0
            -   UnaryOpUGen(HANNING_WINDOW).kr:
                    source: Line.kr[0]
            -   BinaryOpUGen(MULTIPLICATION).kr/0:
                    left: UnaryOpUGen(HANNING_WINDOW).kr[0]
                    right: Control.kr[1:level]
            -   BinaryOpUGen(MULTIPLICATION).kr/1:
                    left: BinaryOpUGen(MULTIPLICATION).kr/0[0]
                    right: Linen.kr[0]
            -   SinOsc.ar:
                    frequency: 440.0
                    phase: 0.0
            -   XOut.ar:
                    bus: Control.ir[1:out]
                    crossfade: BinaryOpUGen(MULTIPLICATION).kr/1[0]
                    source[0]: SinOsc.ar[0]
        """
    )



import os
import pathlib
import platform
from pathlib import Path

import uqbar.io


def _fallback_sclang_path():
    paths = []
    system = platform.system()
    if system == "Linux":
        paths.extend([Path("/usr/bin/sclang"), Path("/usr/local/bin/sclang")])
    elif system == "Darwin":
        paths.append(Path("/Applications/SuperCollider.app/Contents/MacOS/sclang"))
    elif system == "Windows":
        paths.extend(Path(r"C:\Program Files").glob(r"SuperCollider*\sclang.exe"))
    for path in paths:
        if path.exists():
            return path
    return None


def find(sclang_path=None):
    """
    Find the ``sclang`` executable.

    The following paths, if defined, will be searched (prioritised as ordered):

    1. The absolute path ``sclang_path``
    2. The environment variable ``SCLANG_PATH``
    3. The user's ``PATH``
    4. Common installation directories of the SuperCollider application.

    Returns a path to the ``sclang`` executable. Raises ``RuntimeError`` if no path is
    found.
    """
    sclang_path = pathlib.Path(sclang_path or os.environ.get("SCLANG_PATH") or "sclang")
    if sclang_path.is_absolute() and uqbar.io.find_executable(sclang_path):
        return sclang_path
    sclang_path_candidates = uqbar.io.find_executable(sclang_path.name)
    if sclang_path_candidates:
        return pathlib.Path(sclang_path_candidates[0])
    fallback_path = _fallback_sclang_path()
    if fallback_path is not None:
        return fallback_path
    raise RuntimeError("Failed to locate sclang")



#!/usr/bin/env python
import pathlib
import platform

from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup


def read_version():
    root_path = pathlib.Path(__file__).parent
    version_path = root_path / "supriya" / "_version.py"
    with version_path.open() as file_pointer:
        file_contents = file_pointer.read()
    local_dict = {}
    exec(file_contents, None, local_dict)
    return local_dict["__version__"]


extensions = [
    Extension(
        "supriya.utils._intervals",
        language="c",
        sources=["supriya/utils/_intervals.pyx"],
    )
]

if platform.system() != "Windows":
    extensions.append(
        Extension(
            "supriya.contexts.shm",
            include_dirs=[
                "vendor",
                "vendor/TLSF-2.4.6/src",
                "vendor/supercollider/common",
            ],
            language="c++",
            libraries=["rt"] if platform.system() == "Linux" else [],
            sources=["supriya/contexts/shm.pyx"],
        )
    )

if __name__ == "__main__":
    setup(
        ext_modules=cythonize(extensions),
        packages=find_packages(include=["supriya", "supriya.*"])
        + ["supriya.assets.audio", "supriya.assets.audio.birds"],
        version=read_version(),
        package_data={"supriya": ["py.typed"]},
    )



from ...ugens import (
    HPF,
    LPF,
    CompanderD,
    In,
    Limiter,
    ReplaceOut,
    Sum4,
    SynthDef,
    SynthDefBuilder,
)


def _make_synthdef(channel_count: int = 2) -> SynthDef:
    with SynthDefBuilder(
        frequency_1=200,
        frequency_2=2000,
        frequency_3=5000,
        band_1_clamp_time=0.01,
        band_1_postgain=0,
        band_1_pregain=0,
        band_1_relax_time=0.1,
        band_1_slope_above=0.5,
        band_1_slope_below=1.0,
        band_1_threshold=-12,
        band_2_clamp_time=0.01,
        band_2_postgain=0,
        band_2_pregain=0,
        band_2_relax_time=0.1,
        band_2_slope_above=0.5,
        band_2_slope_below=1.0,
        band_2_threshold=-12,
        band_3_clamp_time=0.01,
        band_3_postgain=0,
        band_3_pregain=0,
        band_3_relax_time=0.1,
        band_3_slope_above=0.5,
        band_3_slope_below=1.0,
        band_3_threshold=-12,
        band_4_clamp_time=0.01,
        band_4_postgain=0,
        band_4_pregain=0,
        band_4_relax_time=0.1,
        band_4_slope_above=0.5,
        band_4_slope_below=1.0,
        band_4_threshold=-12,
        pregain=0,
        postgain=0,
        in_=0,
        out=0,
        limiter_lookahead=0.01,
    ) as builder:
        source = In.ar(bus=builder["in_"], channel_count=channel_count)
        source *= builder["pregain"].db_to_amplitude()
        band_1 = LPF.ar(frequency=builder["frequency_1"], source=source)
        band_4 = HPF.ar(frequency=builder["frequency_3"], source=source)
        center = source - band_1 - band_4
        band_2 = LPF.ar(frequency=builder["frequency_2"], source=center)
        band_3 = HPF.ar(frequency=builder["frequency_2"], source=center)
        band_1 = CompanderD.ar(
            clamp_time=builder["band_1_clamp_time"],
            relax_time=builder["band_1_relax_time"],
            slope_above=builder["band_1_slope_above"],
            slope_below=builder["band_1_slope_below"],
            source=band_1 * builder["band_1_pregain"].db_to_amplitude(),
            threshold=builder["band_1_threshold"].db_to_amplitude(),
        )
        band_2 = CompanderD.ar(
            clamp_time=builder["band_2_clamp_time"],
            relax_time=builder["band_2_relax_time"],
            slope_above=builder["band_2_slope_above"],
            slope_below=builder["band_2_slope_below"],
            source=band_2 * builder["band_2_pregain"].db_to_amplitude(),
            threshold=builder["band_2_threshold"].db_to_amplitude(),
        )
        band_3 = CompanderD.ar(
            clamp_time=builder["band_3_clamp_time"],
            relax_time=builder["band_3_relax_time"],
            slope_above=builder["band_3_slope_above"],
            slope_below=builder["band_3_slope_below"],
            source=band_3 * builder["band_3_pregain"].db_to_amplitude(),
            threshold=builder["band_3_threshold"].db_to_amplitude(),
        )
        band_4 = CompanderD.ar(
            clamp_time=builder["band_4_clamp_time"],
            relax_time=builder["band_4_relax_time"],
            slope_above=builder["band_4_slope_above"],
            slope_below=builder["band_4_slope_below"],
            source=band_4 * builder["band_4_pregain"].db_to_amplitude(),
            threshold=builder["band_4_threshold"].db_to_amplitude(),
        )
        band_1 *= builder["band_1_postgain"].db_to_amplitude()
        band_2 *= builder["band_2_postgain"].db_to_amplitude()
        band_3 *= builder["band_3_postgain"].db_to_amplitude()
        band_4 *= builder["band_4_postgain"].db_to_amplitude()
        source = Sum4.new(
            input_one=band_1, input_two=band_2, input_three=band_3, input_four=band_4
        )
        source *= builder["postgain"].db_to_amplitude()
        source = Limiter.ar(source=source, duration=builder["limiter_lookahead"])
        ReplaceOut.ar(bus=builder["out"], source=source)
    return builder.build()


multiband_compressor = _make_synthdef(channel_count=2)

__all__ = ["multiband_compressor"]



"""
Exceptions.
"""


class AlreadyAllocated(Exception):
    pass


class NotAllocated(Exception):
    pass


class BufferAlreadyAllocated(AlreadyAllocated):
    pass


class BufferNotAllocated(NotAllocated):
    pass


class BusAlreadyAllocated(AlreadyAllocated):
    pass


class BusNotAllocated(NotAllocated):
    pass


class IncompatibleRate(Exception):
    pass


class NodeAlreadyAllocated(AlreadyAllocated):
    pass


class NodeNotAllocated(NotAllocated):
    pass


class NonrealtimeOutputMissing(Exception):
    pass


class NonrealtimeRenderError(Exception):
    pass


class RequestTimeout(Exception):
    pass


class ServerCannotBoot(Exception):
    pass


class ServerOnline(Exception):
    pass


class ServerOffline(Exception):
    pass


class OwnedServerShutdown(Exception):
    pass


class UnownedServerShutdown(Exception):
    pass


class TooManyClients(Exception):
    pass



import pytest
from uqbar.iterables import group_by_count, nwise

from supriya.patterns import ShufflePattern


@pytest.mark.parametrize(
    "sequence, iterations, forbid_repetitions, stride, is_infinite",
    [
        ([1], 1, False, 1, False),
        ([1], None, False, 1, True),
        ([1, 2, 3], 1, False, 3, False),
        ([1, 2, 3], 2, False, 3, False),
        ([1, 2, 3], None, False, 3, True),
    ],
)
def test(sequence, iterations, forbid_repetitions, stride, is_infinite):
    pattern = ShufflePattern(
        sequence, iterations=iterations, forbid_repetitions=forbid_repetitions
    )
    assert pattern.is_infinite == is_infinite
    iterator = iter(pattern)
    ceased = True
    actual = []
    for _ in range(1000):
        try:
            actual.append(next(iterator))
        except StopIteration:
            break
    else:
        ceased = False
    if is_infinite:
        assert not ceased
    if forbid_repetitions:
        for a, b in nwise(actual):
            assert a != b
    chunks = list(tuple(chunk) for chunk in group_by_count(actual, stride))
    if is_infinite and len(chunks[-1]) != stride:
        chunks.pop()  # If final group is short due to 1000 iterations, discard
    for chunk in chunks:
        assert len(chunk) == len(set(chunk)), chunk
    assert len(set(tuple(sorted(chunk)) for chunk in chunks)) == 1



import uuid
from unittest.mock import Mock, call

import pytest

from supriya import AddAction
from supriya.contexts import Group, Server
from supriya.patterns.events import GroupAllocateEvent, Priority

id_ = uuid.uuid4()


@pytest.mark.parametrize(
    "event, offset, expected",
    [(GroupAllocateEvent(id_), 0.0, [(0.0, Priority.START, GroupAllocateEvent(id_))])],
)
def test_expand(event, offset, expected):
    actual = event.expand(offset)
    assert actual == expected


def test_perform():
    context = Server().boot()
    spy = Mock(wraps=context)
    proxy_mapping = {}
    notes_mapping = {}
    # Allocate
    event = GroupAllocateEvent(id_)
    with context.at():
        event.perform(
            spy,
            proxy_mapping,
            current_offset=0.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    assert proxy_mapping == {id_: Group(context=context, id_=1000)}
    assert notes_mapping == {}
    assert spy.mock_calls == [
        call.add_group(add_action=AddAction.ADD_TO_HEAD, target_node=None)
    ]



import pytest

from supriya.patterns import ChoicePattern, SeedPattern, SequencePattern
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, pattern, expected, is_infinite",
    [
        (None, SequencePattern([1, 2, 3, 4]), [1, 2, 3, 4], False),
        (1, SequencePattern([1, 2, 3, 4]), [1], False),
    ],
)
def test(stop_at, pattern, expected, is_infinite):
    pattern = SeedPattern(pattern)
    run_pattern_test(pattern, expected, is_infinite, stop_at)


def test_random():
    pattern = ChoicePattern([1, 2, 3])
    assert len(set(tuple(pattern) for _ in range(10))) > 1
    pattern = SeedPattern(ChoicePattern([1, 2, 3]))
    assert len(set(tuple(pattern) for _ in range(10))) == 1



import bisect
from typing import Dict, Generator, List, Optional, Sequence, SupportsInt, Tuple
from uuid import UUID, uuid4

from uqbar.objects import get_vars, new

from supriya.assets import synthdefs
from supriya.enums import CalculationRate

from ..typing import CalculationRateLike
from ..ugens import SynthDef
from .events import (
    BusAllocateEvent,
    BusFreeEvent,
    CompositeEvent,
    Event,
    GroupAllocateEvent,
    NodeFreeEvent,
    NullEvent,
    SynthAllocateEvent,
)
from .patterns import Pattern

UUIDDict = Dict[str, UUID]


class BusPattern(Pattern[Event]):
    """
    Peform node events from a pattern into a dynamically allocated bus and group.
    """

    ### INITIALIZER ###

    def __init__(
        self,
        pattern: Pattern[Event],
        calculation_rate: CalculationRateLike = "audio",
        channel_count: int = 1,
        release_time: float = 0.25,
    ) -> None:
        self._pattern = pattern
        self._calculation_rate: CalculationRate = CalculationRate.from_expr(
            calculation_rate
        )
        self._channel_count = channel_count
        self._release_time = release_time

    ### PRIVATE METHODS ###

    def _adjust(self, expr: Event, state: Optional[UUIDDict] = None) -> Event:
        if state is None:
            raise RuntimeError
        args, _, kwargs = get_vars(expr)
        updates = {}
        if hasattr(expr, "target_node") and expr.target_node is None:
            updates["target_node"] = state["group"]
        if hasattr(expr, "synthdef"):
            synthdef = getattr(expr, "synthdef") or synthdefs.default
            parameter_names = synthdef.parameters
            for name in ("in_", "out"):
                if name in parameter_names and kwargs.get(name) is None:
                    updates[name] = state["bus"]
        if updates:
            return new(expr, **updates)
        return expr

    def _iterate(
        self, state: Optional[UUIDDict] = None
    ) -> Generator[Event, bool, None]:
        return iter(self._pattern)

    def _setup_peripherals(self, state: Optional[UUIDDict]) -> Tuple[Event, Event]:
        if state is None:
            raise RuntimeError
        rate = self._calculation_rate.name.lower()
        link_synthdef_name = f"system_link_{rate}_{self._channel_count}"
        starts = [
            BusAllocateEvent(
                calculation_rate=self._calculation_rate,
                channel_count=self._channel_count,
                id_=state["bus"],
            ),
            GroupAllocateEvent(id_=state["group"]),
            SynthAllocateEvent(
                add_action="ADD_AFTER",
                amplitude=1.0,
                fade_time=self._release_time,
                in_=state["bus"],
                synthdef=getattr(synthdefs, link_synthdef_name),
                target_node=state["group"],
                id_=state["link"],
            ),
        ]
        stops = [
            NodeFreeEvent(id_=state["link"]),
            NodeFreeEvent(id_=state["group"]),
            BusFreeEvent(id_=state["bus"]),
        ]
        if self._release_time:
            stops.insert(1, NullEvent(delta=self._release_time))
        return CompositeEvent(starts), CompositeEvent(stops)

    def _setup_state(self) -> UUIDDict:
        return {"bus": uuid4(), "link": uuid4(), "group": uuid4()}

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self) -> bool:
        return self._pattern.is_infinite


class FxPattern(Pattern[Event]):
    """
    Add a synth to the tail of the nodes performed by a pattern.
    """

    ### INITIALIZER ###

    def __init__(
        self,
        pattern: Pattern[Event],
        synthdef: SynthDef,
        release_time: float = 0.25,
        **kwargs,
    ) -> None:
        self._pattern = pattern
        self._release_time = release_time
        self._synthdef = synthdef
        self._kwargs = kwargs

    ### PRIVATE METHODS ###

    def _iterate(
        self, state: Optional[UUIDDict] = None
    ) -> Generator[Event, bool, None]:
        return iter(self._pattern)

    def _setup_peripherals(
        self, state: Optional[UUIDDict] = None
    ) -> Tuple[Event, Event]:
        if state is None:
            raise RuntimeError
        starts: List[Event] = [
            SynthAllocateEvent(
                add_action="ADD_TO_TAIL",
                synthdef=self._synthdef,
                id_=state["synth"],
                **self._kwargs,
            )
        ]
        stops: List[Event] = [NodeFreeEvent(id_=state["synth"])]
        if self._release_time:
            stops.insert(0, NullEvent(delta=self._release_time))
        return CompositeEvent(starts), CompositeEvent(stops)

    def _setup_state(self) -> UUIDDict:
        return {"synth": uuid4()}

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self) -> bool:
        return self._pattern.is_infinite


class GroupPattern(Pattern[Event]):
    """
    Perform node events from a pattern into a dynamically allocated group.
    """

    ### INITIALIZER ###

    def __init__(self, pattern: Pattern[Event], release_time: float = 0.25) -> None:
        self._pattern = pattern
        self._release_time = release_time

    ### PRIVATE METHODS ###

    def _adjust(self, expr: Event, state: Optional[UUIDDict] = None) -> Event:
        if state is None:
            raise RuntimeError
        updates = {}
        if hasattr(expr, "target_node") and expr.target_node is None:
            updates["target_node"] = state["group"]
        if updates:
            return new(expr, **updates)
        return expr

    def _iterate(
        self, state: Optional[UUIDDict] = None
    ) -> Generator[Event, bool, None]:
        return iter(self._pattern)

    def _setup_peripherals(self, state: Optional[UUIDDict]) -> Tuple[Event, Event]:
        if state is None:
            raise RuntimeError
        starts: List[Event] = [
            GroupAllocateEvent(add_action="ADD_TO_HEAD", id_=state["group"])
        ]
        stops: List[Event] = [NodeFreeEvent(id_=state["group"])]
        if self._release_time:
            stops.insert(0, NullEvent(delta=self._release_time))
        return CompositeEvent(starts), CompositeEvent(stops)

    def _setup_state(self) -> UUIDDict:
        return {"group": uuid4()}

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self) -> bool:
        return self._pattern.is_infinite


class ParallelPattern(Pattern[Event]):
    """
    Perform patterns simultaneously in parallel.
    """

    ### INITIALIZER ###

    def __init__(self, patterns: Sequence[Pattern[Event]]) -> None:
        self._patterns = tuple(patterns)

    ### PRIVATE METHODS ###

    def _iterate(
        self, state: Optional[Dict[str, UUID]] = None
    ) -> Generator[Event, bool, None]:
        should_stop = False
        iterators = []
        for index, pattern in enumerate(self._patterns):
            iterators.append((0.0, index, iter(pattern)))
        while iterators:
            grouping_offset = iterators[0][0]
            events = []
            while iterators and iterators[0][0] == grouping_offset:
                offset, index, iterator = iterators.pop(0)
                try:
                    if should_stop:
                        event = iterator.send(should_stop)
                    else:
                        event = next(iterator)
                    events.append(event)
                    triple = (offset + event.delta, index, iterator)
                    insert_index = bisect.bisect_left(iterators, triple)
                    iterators.insert(insert_index, triple)
                except StopIteration:
                    pass
            if events:
                if iterators:
                    delta = iterators[0][0] - grouping_offset
                else:
                    delta = max(event.delta for event in events)
                if len(events) == 1:
                    sent = yield new(events[0], delta=delta)
                elif len(events) > 1:
                    sent = yield CompositeEvent(
                        [new(x, delta=0.0) for x in events], delta=delta
                    )
                if sent:
                    should_stop = True

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self) -> bool:
        return any(pattern.is_infinite for pattern in self._patterns)


class PinPattern(Pattern[Event]):
    """
    Utility pattern for assigning an explicit target bus and/or target node to NodeEvents.

    Used internally by pattern players.
    """

    ### INITIALIZER ###

    def __init__(
        self,
        pattern: Pattern[Event],
        *,
        target_bus: Optional[SupportsInt] = None,
        target_node: Optional[SupportsInt] = None,
    ) -> None:
        self._pattern = pattern
        self._target_bus = target_bus
        self._target_node = target_node

    def _adjust(self, expr: Event, state: Optional[UUIDDict] = None) -> Event:
        args, _, kwargs = get_vars(expr)
        updates = {}
        if self._target_node is not None and hasattr(expr, "target_node"):
            updates["target_node"] = expr.target_node or self._target_node
        if self._target_bus is not None and hasattr(expr, "synthdef"):
            synthdef = getattr(expr, "synthdef") or synthdefs.default
            parameter_names = synthdef.parameters
            for name in ("in_", "out"):
                if name in parameter_names and kwargs.get(name) is None:
                    updates[name] = self._target_bus
        if updates:
            return new(expr, **updates)
        return expr

    ### PRIVATE METHODS ###

    def _iterate(
        self, state: Optional[Dict[str, UUID]] = None
    ) -> Generator[Event, bool, None]:
        return iter(self._pattern)

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self) -> bool:
        return self._pattern.is_infinite



from .core import UGen, param, ugen


@ugen(ar=True, is_multichannel=True, has_done_flag=True)
class DiskIn(UGen):
    """
    Streams in audio from a file.

    ::

        >>> buffer_id = 23
        >>> disk_in = supriya.ugens.DiskIn.ar(
        ...     buffer_id=buffer_id,
        ...     channel_count=2,
        ...     loop=0,
        ... )
        >>> disk_in
        <DiskIn.ar()>
    """

    buffer_id = param()
    loop = param(0)


@ugen(ar=True)
class DiskOut(UGen):
    """
    Records to a soundfile to disk.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.SinOsc.ar(frequency=[440, 442])
        >>> disk_out = supriya.ugens.DiskOut.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        >>> disk_out
        <DiskOut.ar()[0]>
    """

    buffer_id = param()
    source = param(unexpanded=True)


@ugen(ar=True, is_multichannel=True, has_done_flag=True)
class VDiskIn(UGen):
    """
    Streams in audio from a file, with variable rate.

    ::

        >>> buffer_id = 23
        >>> vdisk_in = supriya.ugens.VDiskIn.ar(
        ...     buffer_id=buffer_id,
        ...     channel_count=2,
        ...     loop=0,
        ...     rate=1,
        ...     send_id=0,
        ... )
        >>> vdisk_in
        <VDiskIn.ar()>
    """

    buffer_id = param()
    rate = param(1.0)
    loop = param(0)
    send_id = param(0)



import os
import pathlib
import stat
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Dict, List

import pytest

from supriya import scsynth


@pytest.fixture
def mock_env_scsynth_path(monkeypatch):
    monkeypatch.delenv(scsynth.ENVAR_SERVER_EXECUTABLE, raising=False)
    monkeypatch.setenv("PATH", "")


def test_find_argument(mock_env_scsynth_path):
    with NamedTemporaryFile() as tmp:
        expected = pathlib.Path(tmp.name).absolute()
        expected.chmod(expected.stat().st_mode | stat.S_IEXEC)
        got = scsynth.find(expected)
        assert got == expected


def test_find_env_var(mock_env_scsynth_path, monkeypatch):
    with NamedTemporaryFile() as tmp:
        expected = pathlib.Path(tmp.name).absolute()
        expected.chmod(expected.stat().st_mode | stat.S_IEXEC)
        monkeypatch.setenv(scsynth.ENVAR_SERVER_EXECUTABLE, str(expected))
        got = scsynth.find()
        assert got == expected


def test_find_on_path(mock_env_scsynth_path, monkeypatch):
    with TemporaryDirectory() as tmp_dir:
        scsynth_path = pathlib.Path(tmp_dir) / "scsynth"
        with open(scsynth_path, "w"):
            scsynth_path.chmod(scsynth_path.stat().st_mode | stat.S_IEXEC)
            monkeypatch.setenv("PATH", os.pathsep + tmp_dir)
            got = scsynth.find()
            expected = scsynth_path.resolve().absolute()
            assert got == expected


@pytest.mark.parametrize(
    "kwargs, expected",
    [
        (
            {},
            [
                "/path/to/scsynth",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
        # only input device defined
        (
            {"input_device": "Device X"},
            [
                "/path/to/scsynth",
                "-H",
                "Device X",
                "",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
        # only output device defined
        (
            {"output_device": "Device Y"},
            [
                "/path/to/scsynth",
                "-H",
                "",
                "Device Y",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
        # input and output devices defined, but different
        (
            {"input_device": "Device P", "output_device": "Device Q"},
            [
                "/path/to/scsynth",
                "-H",
                "Device P",
                "Device Q",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
        # input and output devices defined, and identical
        (
            {"input_device": "Device Z", "output_device": "Device Z"},
            [
                "/path/to/scsynth",
                "-H",
                "Device Z",
                "-R",
                "0",
                "-l",
                "1",
                "-u",
                "57110",
            ],
        ),
    ],
)
def test_Options(kwargs: Dict, expected: List[str]) -> None:
    options = scsynth.Options(**kwargs)
    actual = list(options)
    actual[0] = "/path/to/scsynth"  # replace to make it portable
    assert actual == expected



import logging
import queue
from typing import Generator, Optional, Tuple

from .asynchronous import AsyncClock
from .bases import BaseClock
from .ephemera import CallbackEvent, ClockContext, Moment, TimeUnit

logger = logging.getLogger(__name__)


class OfflineClock(BaseClock):
    def __init__(self) -> None:
        super().__init__()
        self._generator: Optional[Generator[bool, None, None]] = None

    ### SCHEDULING METHODS ###

    def _perform_callback_event(
        self, event: CallbackEvent, current_moment: Moment, desired_moment: Moment
    ) -> None:
        logger.debug(
            f"[{self.name}] ... ... Performing {event.procedure} at "
            f"{desired_moment.seconds - self._state.initial_seconds}:s / "
            f"{desired_moment.offset}:o"
        )
        context = ClockContext(current_moment, desired_moment, event)
        args = event.args or ()
        kwargs = event.kwargs or {}
        result = event.procedure(context, *args, **kwargs)
        if isinstance(result, float) or result is None:
            delta, time_unit = result, TimeUnit.BEATS
        else:
            delta, time_unit = result
        self._process_callback_event_result(desired_moment, event, delta, time_unit)

    def _run(self, offline: bool = False) -> Generator[bool, None, None]:
        logger.debug(f"[{self.name}] Thread start")
        self._process_command_deque(first_run=True)
        while self._is_running and self._event_queue.qsize():
            logger.debug(f"[{self.name}] Loop start")
            if not self._wait_for_queue():
                return
            try:
                if (current_moment := self._wait_for_moment()) is None:
                    return
            except queue.Empty:
                continue
            current_moment = self._perform_events(current_moment)
            self._state = self._state._replace(
                previous_seconds=current_moment.seconds,
                previous_offset=current_moment.offset,
            )
        logger.debug(f"[{self.name}] Terminating")
        yield False
        self._stop()

    def _wait_for_moment(self, offline: bool = False) -> Optional[Moment]:
        current_time = self._event_queue.peek().seconds
        return self._seconds_to_moment(current_time)

    def _wait_for_queue(self, offline: bool = False) -> bool:
        logger.debug(f"[{self.name}] ... Waiting for events")
        self._process_command_deque()
        return True

    ### PUBLIC METHODS ###

    def get_current_time(self) -> float:
        if not self._is_running:
            return 0.0
        return self._state.previous_seconds

    def start(
        self,
        initial_time: Optional[float] = None,
        initial_offset: float = 0.0,
        initial_measure: int = 1,
        beats_per_minute: Optional[float] = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ) -> None:
        self._start(
            initial_time=initial_time,
            initial_offset=initial_offset,
            initial_measure=initial_measure,
            beats_per_minute=beats_per_minute,
            time_signature=time_signature,
        )
        self._generator = self._run()
        if not next(self._generator):
            self._stop()

    def stop(self) -> None:
        if not self._stop():
            return
        if self._generator is not None:
            while True:
                try:
                    next(self._generator)
                except StopIteration:
                    pass


class AsyncOfflineClock(AsyncClock):
    async def _run_async(self, offline: bool = False) -> None:
        logger.debug(f"[{self.name}] Coroutine start")
        self._process_command_deque(first_run=True)
        while self._is_running and self._event_queue.qsize():
            logger.debug(f"[{self.name}] Loop start")
            if not await self._wait_for_queue_async():
                return
            try:
                if (current_moment := await self._wait_for_moment_async()) is None:
                    return
            except queue.Empty:
                continue
            current_moment = await self._perform_events_async(current_moment)
            self._state = self._state._replace(
                previous_seconds=current_moment.seconds,
                previous_offset=current_moment.offset,
            )
        logger.debug(f"[{self.name}] Coroutine terminating")
        self._stop()

    async def _wait_for_event_async(self, sleep_time: float) -> None:
        pass

    async def _wait_for_moment_async(self, offline: bool = False) -> Optional[Moment]:
        current_time = self._event_queue.peek().seconds
        self._process_command_deque()
        self._event.clear()
        return self._seconds_to_moment(current_time)

    async def _wait_for_queue_async(self, offline: bool = False) -> bool:
        logger.debug(f"[{self.name}] ... Waiting for events")
        self._process_command_deque()
        self._event.clear()
        return True

    def get_current_time(self) -> float:
        if not self._is_running:
            return 0.0
        return self._state.previous_seconds



import logging
import os
import pathlib
import platform
import shutil

import pytest
import uqbar.io


@pytest.fixture
def rm_dirs(app):
    for path in [pathlib.Path(app.doctreedir), pathlib.Path(app.outdir) / "_images"]:
        if path.exists():
            shutil.rmtree(path)
    yield


@pytest.mark.skipif(bool(os.environ.get("CI")), reason="Import breaks under GHA")
@pytest.mark.sphinx("html", testroot="book")
def test_sphinx_book_html(caplog, app, status, warning, rm_dirs):
    caplog.set_level(logging.INFO)
    app.build()
    for root, paths, dirs in uqbar.io.walk(app.outdir):
        for path in paths:
            print(path)
    assert not warning.getvalue().strip()
    image_path = pathlib.Path(app.outdir) / "_images"
    expected_file_names = [
        "score-c073aea37cd922f2fe26dbd47dfecee37e1cb2e6cdd06e86908b090add6e2c7e.aiff",
        "score-c073aea37cd922f2fe26dbd47dfecee37e1cb2e6cdd06e86908b090add6e2c7e.osc",
    ]
    if platform.system() != "Windows":
        expected_file_names.extend(
            [
                "say-3d7f815142f4edadfca73bd01c264223de1e4f5fd5e50a4c9e6f917eddeddba6.aiff",
                "say-3d7f815142f4edadfca73bd01c264223de1e4f5fd5e50a4c9e6f917eddeddba6.mp3",
            ]
        )
    # Haven't found a /simple/ cross-platforn solution for detecting audio device SR
    expected_44100_file_names = expected_file_names + [
        "audio-08abe38d842cbaa19789618fe4675f1cf64de0eb6f9ab7ebd2165c078ce31429.mp3",
        "audio-08abe38d842cbaa19789618fe4675f1cf64de0eb6f9ab7ebd2165c078ce31429.wav",
        "plot-397eb6446e5486ac9137cb98affdda8577148ae41ef7857807f53be0793bc74a.svg",
    ]
    expected_48000_file_names = expected_file_names + [
        "audio-4f0fd44621b74146c936fab67a7544438ddb60abe59b506082268778ec2e285f.mp3",
        "audio-4f0fd44621b74146c936fab67a7544438ddb60abe59b506082268778ec2e285f.wav",
        "plot-6e8cafcdb775004ffba3051b743f9d5a539d541c6601f723ecfcc3145f0217b4.svg",
    ]
    actual_file_names = sorted(path.name for path in image_path.iterdir())
    print(actual_file_names)
    assert all(
        file_name in actual_file_names for file_name in expected_44100_file_names
    ) or all(file_name in actual_file_names for file_name in expected_48000_file_names)
    # audio and plot names are not stable across platforms
    audio_mp3_paths = list(image_path.glob("audio-*.mp3"))
    audio_wav_paths = list(image_path.glob("audio-*.wav"))
    plot_svg_paths = list(image_path.glob("plot-*.svg"))
    assert len(audio_mp3_paths) == 1
    assert len(audio_wav_paths) == 1
    assert len(plot_svg_paths) == 1


@pytest.mark.sphinx("text", testroot="book")
def test_sphinx_book_text(app, status, warning, rm_dirs):
    app.build()
    assert not warning.getvalue().strip()



import pytest
from uqbar.iterables import nwise

from supriya.patterns import ChoicePattern, SequencePattern


@pytest.mark.parametrize(
    "sequence, iterations, forbid_repetitions, weights, is_infinite",
    [
        ([1, 2, 3], 1, False, None, False),
        ([1, 2, 3], 1, True, None, False),
        ([1, 2, 3], None, False, None, True),
        ([1, 2, 3], None, True, None, True),
        ([1, 2, 3], None, True, [1, 2, 1], True),
        (
            [SequencePattern(["a", "b"]), SequencePattern(["c", "d"])],
            None,
            False,
            None,
            True,
        ),
    ],
)
def test(sequence, iterations, forbid_repetitions, weights, is_infinite):
    pattern = ChoicePattern(
        sequence,
        iterations=iterations,
        forbid_repetitions=forbid_repetitions,
        weights=weights,
    )
    assert pattern.is_infinite == is_infinite
    iterator = iter(pattern)
    ceased = True
    actual = []
    for _ in range(1000):
        try:
            actual.append(next(iterator))
        except StopIteration:
            break
    else:
        ceased = False
    if is_infinite:
        assert not ceased
    if forbid_repetitions:
        for a, b in nwise(actual):
            assert a != b



from typing import Any, Dict, Tuple

from ..enums import CalculationRate, DoneAction
from .core import UGen, param, ugen


@ugen(ar=True, kr=True, is_pure=True)
class COsc(UGen):
    """
    A chorusing wavetable oscillator.

    ::

        >>> cosc = supriya.ugens.COsc.ar(
        ...     beats=0.5,
        ...     buffer_id=23,
        ...     frequency=440,
        ... )
        >>> cosc
        <COsc.ar()[0]>
    """

    buffer_id = param()
    frequency = param(440.0)
    beats = param(0.5)


@ugen(ar=True, kr=True, is_pure=True)
class DegreeToKey(UGen):
    """
    A signal-to-modal-pitch converter.`

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> degree_to_key = supriya.ugens.DegreeToKey.ar(
        ...     buffer_id=23,
        ...     octave=12,
        ...     source=source,
        ... )
        >>> degree_to_key
        <DegreeToKey.ar()[0]>
    """

    buffer_id = param()
    source = param()
    octave = param(12)


@ugen(ar=True, kr=True, is_pure=True)
class Impulse(UGen):
    """
    A non-band-limited single-sample impulse generator unit generator.

    ::

        >>> supriya.ugens.Impulse.ar()
        <Impulse.ar()[0]>
    """

    frequency = param(440.0)
    phase = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class Index(UGen):
    """
    A clipping buffer indexer.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> index = supriya.ugens.Index.ar(
        ...     buffer_id=23,
        ...     source=source,
        ... )
        >>> index
        <Index.ar()[0]>
    """

    buffer_id = param()
    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class LFCub(UGen):
    """
    A sine-like oscillator unit generator.

    ::

        >>> supriya.ugens.LFCub.ar()
        <LFCub.ar()[0]>
    """

    frequency = param(440.0)
    initial_phase = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class LFGauss(UGen):
    """
    A non-band-limited gaussian function oscillator.

    ::

        >>> supriya.ugens.LFGauss.ar()
        <LFGauss.ar()[0]>
    """

    duration = param(1)
    width = param(0.1)
    initial_phase = param(0)
    loop = param(1)
    done_action = param(0)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs,
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        return calculation_rate, {
            **kwargs,
            "done_action": DoneAction.from_expr(int(kwargs["done_action"])),
        }


@ugen(ar=True, kr=True, is_pure=True)
class LFPar(UGen):
    """
    A parabolic oscillator unit generator.

    ::

        >>> supriya.ugens.LFPar.ar()
        <LFPar.ar()[0]>
    """

    frequency = param(440.0)
    initial_phase = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class LFPulse(UGen):
    """
    A non-band-limited pulse oscillator.

    ::

        >>> supriya.ugens.LFPulse.ar()
        <LFPulse.ar()[0]>
    """

    frequency = param(440.0)
    initial_phase = param(0.0)
    width = param(0.5)


@ugen(ar=True, kr=True, is_pure=True)
class LFSaw(UGen):
    """
    A non-band-limited sawtooth oscillator unit generator.

    ::

        >>> supriya.ugens.LFSaw.ar()
        <LFSaw.ar()[0]>
    """

    frequency = param(440.0)
    initial_phase = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class LFTri(UGen):
    """
    A non-band-limited triangle oscillator unit generator.

    ::

        >>> supriya.ugens.LFTri.ar()
        <LFTri.ar()[0]>
    """

    frequency = param(440.0)
    initial_phase = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class Osc(UGen):
    """
    An interpolating wavetable oscillator.
    """

    buffer_id = param()
    frequency = param(440.0)
    initial_phase = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class OscN(UGen):
    """
    A non-interpolating wavetable oscillator.
    """

    buffer_id = param()
    frequency = param(440.0)
    initial_phase = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class Select(UGen):
    """
    A signal selector.

    ::

        >>> sources = supriya.ugens.In.ar(bus=0, channel_count=8)
        >>> selector = supriya.ugens.Phasor.kr() * 8
        >>> select = supriya.ugens.Select.ar(
        ...     sources=sources,
        ...     selector=selector,
        ... )
        >>> select
        <Select.ar()[0]>
    """

    selector = param()
    sources = param(unexpanded=True)


@ugen(ar=True, kr=True, is_pure=True)
class SinOsc(UGen):
    """
    A sinusoid oscillator unit generator.

    ::

        >>> supriya.ugens.SinOsc.ar()
        <SinOsc.ar()[0]>

    ::

        >>> print(_)
        synthdef:
            name: ...
            ugens:
            -   SinOsc.ar:
                    frequency: 440.0
                    phase: 0.0
    """

    frequency = param(440.0)
    phase = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class SyncSaw(UGen):
    """
    A sawtooth wave that is hard synched to a fundamental pitch.

    ::

        >>> sync_saw = supriya.ugens.SyncSaw.ar(
        ...     saw_frequency=440,
        ...     sync_frequency=440,
        ... )
        >>> sync_saw
        <SyncSaw.ar()[0]>
    """

    sync_frequency = param(440.0)
    saw_frequency = param(440.0)


@ugen(ar=True, kr=True, is_pure=True)
class VOsc(UGen):
    """
    A wavetable lookup oscillator which can be swept smoothly across wavetables.

    ::

        >>> vosc = supriya.ugens.VOsc.ar(
        ...     buffer_id=supriya.ugens.MouseX.kr(minimum=0, maximum=7),
        ...     frequency=440,
        ...     phase=0,
        ... )
        >>> vosc
        <VOsc.ar()[0]>
    """

    buffer_id = param()
    frequency = param(440.0)
    phase = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class VOsc3(UGen):
    """
    A wavetable lookup oscillator which can be swept smoothly across wavetables.

    ::

        >>> vosc_3 = supriya.ugens.VOsc3.ar(
        ...     buffer_id=supriya.ugens.MouseX.kr(minimum=0, maximum=7),
        ...     freq_1=110,
        ...     freq_2=220,
        ...     freq_3=440,
        ... )
        >>> vosc_3
        <VOsc3.ar()[0]>
    """

    buffer_id = param()
    freq_1 = param(110.0)
    freq_2 = param(220.0)
    freq_3 = param(440.0)


@ugen(ar=True, kr=True, is_pure=True)
class VarSaw(UGen):
    """
    A sawtooth-triangle oscillator with variable duty.

    ::

        >>> supriya.ugens.VarSaw.ar()
        <VarSaw.ar()[0]>
    """

    frequency = param(440.0)
    initial_phase = param(0.0)
    width = param(0.5)


@ugen(ar=True, kr=True, is_pure=True)
class Vibrato(UGen):
    """
    Vibrato is a slow frequency modulation.

    ::

        >>> vibrato = supriya.ugens.Vibrato.ar(
        ...     delay=0,
        ...     depth=0.02,
        ...     depth_variation=0.1,
        ...     frequency=440,
        ...     initial_phase=0,
        ...     onset=0,
        ...     rate=6,
        ...     rate_variation=0.04,
        ... )
        >>> vibrato
        <Vibrato.ar()[0]>
    """

    frequency = param(440)
    rate = param(6)
    depth = param(0.02)
    delay = param(0)
    onset = param(0)
    rate_variation = param(0.04)
    depth_variation = param(0.1)
    initial_phase = param(0)


@ugen(ar=True, kr=True, is_pure=True)
class WrapIndex(UGen):
    """
    A wrapping buffer indexer.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> wrap_index = supriya.ugens.WrapIndex.ar(
        ...     buffer_id=23,
        ...     source=source,
        ... )
        >>> wrap_index
        <WrapIndex.ar()[0]>
    """

    buffer_id = param()
    source = param()



import itertools
from typing import (
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

T = TypeVar("T")
IT = Iterable[Union[T, "IT"]]


class Expander(Generic[T]):
    def __call__(
        self,
        mapping: Dict[str, Union[T, Sequence[T]]],
        unexpanded: Optional[Iterable[str]] = None,
        only: Optional[Iterable[str]] = None,
    ) -> List[Dict[str, Union[T, Sequence[T]]]]:
        only_ = set(only or ())
        unexpanded_ = set(unexpanded or ())
        expanded_mappings = []
        maximum_length = 1
        massaged: Dict[str, Sequence[T]] = {}
        for key, value in mapping.items():
            if only_ and key not in only_:
                continue
            if isinstance(value, Sequence):
                if key not in unexpanded_:
                    maximum_length = max(len(value), maximum_length)
            else:
                value = [value]
            massaged[key] = value
        for i in range(maximum_length):
            expanded_mapping: Dict[str, Union[T, Sequence[T]]] = {}
            for key, value in massaged.items():
                if key in unexpanded_:
                    expanded_mapping[key] = value
                else:
                    expanded_mapping[key] = value[i % len(value)]
            expanded_mappings.append(expanded_mapping)
        return expanded_mappings


def expand(
    mapping: Dict[str, Union[T, Sequence[T]]],
    unexpanded: Optional[Iterable[str]] = None,
    only: Optional[Iterable[str]] = None,
) -> List[Dict[str, Union[T, Sequence[T]]]]:
    return Expander[T]()(mapping, unexpanded, only)


def flatten(
    iterable: IT, terminal_types: Optional[Union[Type, Tuple[Type, ...]]] = None
) -> Generator[T, None, None]:
    for x in iterable:
        if isinstance(x, Iterable) and (
            terminal_types is None or not isinstance(x, terminal_types)
        ):
            yield from flatten(x, terminal_types)
        else:
            yield x


def group_by_count(iterable: Iterable[T], count: int) -> Generator[List[T], None, None]:
    iterator = iter(iterable)
    while True:
        group = list(itertools.islice(iterator, count))
        if not group:
            return
        yield group


def iterate_nwise(
    iterable: Iterable[T], n: int = 2
) -> Generator[Sequence[T], None, None]:
    iterables = itertools.tee(iterable, n)
    temp: List[Iterable[T]] = []
    for idx, it in enumerate(iterables):
        it = itertools.islice(it, idx, None)
        temp.append(it)
    yield from zip(*temp)


def repeat_to_length(iterable: Iterable[T], length: int) -> Generator[T, None, None]:
    for i, x in enumerate(itertools.cycle(iterable)):
        if i >= length:
            break
        yield x


def zip_cycled(*args: Iterable[T]) -> Generator[Iterable[T], None, None]:
    maximum_i = max(len(_) for _ in args) - 1
    cycles = [itertools.cycle(_) for _ in args]
    iterator = enumerate(zip(*cycles))
    for i, result in iterator:
        yield result
        if i == maximum_i:
            break



import asyncio
import base64
import hashlib
import logging
import pathlib
import pickle
import textwrap
import warnings
from typing import Union

import librosa.display
import matplotlib.axes  # noqa
import matplotlib.pyplot as plt
from docutils.nodes import FixedTextElement, General, SkipNode
from uqbar.book.extensions import Extension
from uqbar.strings import normalize

from ..io import Player, Plotter
from ..typing import SupportsRender, SupportsRenderMemo
from . import websafe_audio

logger = logging.getLogger(__name__)


class PlayExtension(Extension):
    template = normalize(
        """
        <audio controls src="{file_path}">
            Your browser does not support the <code>audio</code> element.
        </audio>
        """
    )

    class play_block(General, FixedTextElement):
        pass

    @classmethod
    def setup_console(cls, console, monkeypatch):
        monkeypatch.setattr(
            Player,
            "__call__",
            lambda self: console.push_proxy(cls(self.renderable, self.render_kwargs)),
        )

    @classmethod
    def setup_sphinx(cls, app):
        app.add_node(
            cls.play_block,
            html=[cls.visit_block_html, None],
            latex=[cls.visit_block_latex, None],
            text=[cls.visit_block_text, cls.depart_block_text],
        )

    def __init__(
        self, renderable: Union[SupportsRender, SupportsRenderMemo], render_kwargs
    ) -> None:
        if isinstance(renderable, SupportsRenderMemo):
            self.renderable = pickle.loads(pickle.dumps(renderable.__render_memo__()))
        else:
            self.renderable = pickle.loads(pickle.dumps(renderable))
        self.render_kwargs = pickle.loads(pickle.dumps(render_kwargs))

    def to_docutils(self):
        code = "\n".join(
            textwrap.wrap(
                base64.b64encode(
                    pickle.dumps((self.renderable, self.render_kwargs))
                ).decode()
            )
        )
        node = self.play_block(code, code)
        return [node]

    @classmethod
    def render(cls, node, output_path):
        output_path.mkdir(exist_ok=True)
        renderable, render_kwargs = pickle.loads(
            base64.b64decode("".join(node[0].split()))
        )
        path, _ = asyncio.run(
            renderable.__render__(
                render_directory_path=output_path,
                **render_kwargs,
            )
        )
        return websafe_audio(path)

    @staticmethod
    def visit_block_html(self, node):
        absolute_file_path = PlayExtension.render(
            node, pathlib.Path(self.builder.outdir) / "_images"
        )
        relative_file_path = (
            pathlib.Path(self.builder.imgpath) / absolute_file_path.name
        )
        result = PlayExtension.template.format(file_path=relative_file_path)
        self.body.append(result)
        raise SkipNode


class PlotExtension(Extension):
    template = normalize(
        """
        <object data="{file_path}" type="image/svg+xml"></object>
        """
    )

    class plot_block(General, FixedTextElement):
        pass

    @classmethod
    def setup_console(cls, console, monkeypatch):
        monkeypatch.setattr(
            Plotter, "__call__", lambda self: console.push_proxy(cls(self.plottable))
        )

    @classmethod
    def setup_sphinx(cls, app):
        # Suppress "PySoundFile failed. Trying audioread instead."
        warnings.filterwarnings(
            "ignore", category=UserWarning, module="librosa.core.audio"
        )
        app.add_node(
            cls.plot_block,
            html=[cls.visit_block_html, None],
            latex=[cls.visit_block_latex, None],
            text=[cls.visit_block_text, cls.depart_block_text],
        )

    def __init__(self, plottable, **kwargs):
        self.array, self.sample_rate = plottable.__plot__()

    def to_docutils(self):
        code = "\n".join(
            textwrap.wrap(
                base64.b64encode(pickle.dumps((self.array, self.sample_rate))).decode()
            )
        )
        node = self.plot_block(code, code)
        return [node]

    @classmethod
    def render(cls, node, output_path):
        output_path.mkdir(exist_ok=True)
        array, sample_rate = pickle.loads(base64.b64decode("".join(node[0].split())))
        fig, ax = plt.subplots(nrows=1)
        fig.set_size_inches(8, 2)
        fig.patch.set_alpha(0.0)
        ax.set_facecolor("white")
        ax.xaxis.label.set_color("white")
        ax.tick_params(axis="x", colors="white")
        ax.tick_params(axis="y", colors="white")
        # TODO: Drop color="blue" after upgrading librosa > 0.10.1
        #       https://github.com/librosa/librosa/issues/1763
        librosa.display.waveshow(array, sr=sample_rate, ax=ax, color="blue")
        hexdigest = hashlib.sha256(node[0].encode()).hexdigest()
        file_path = output_path / f"plot-{hexdigest}.svg"
        fig.savefig(file_path, bbox_inches="tight")
        plt.close(fig)
        return file_path

    @staticmethod
    def visit_block_html(self, node):
        absolute_file_path = PlotExtension.render(
            node, pathlib.Path(self.builder.outdir) / "_images"
        )
        relative_file_path = (
            pathlib.Path(self.builder.imgpath) / absolute_file_path.name
        )
        result = PlotExtension.template.format(file_path=relative_file_path)
        self.body.append(result)
        raise SkipNode



import pytest

from supriya.patterns.events import NullEvent


@pytest.mark.parametrize(
    "event, offset, expected",
    [(NullEvent(), 0.0, []), (NullEvent(delta=1.5), 0.0, []), (NullEvent(), 2.5, [])],
)
def test_expand(event, offset, expected):
    actual = event.expand(offset)
    assert actual == expected



import asyncio
import concurrent.futures
from os import PathLike
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Coroutine,
    Dict,
    Optional,
    Protocol,
    SupportsFloat,
    SupportsInt,
    Tuple,
    TypeVar,
    Union,
    runtime_checkable,
)

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias  # noqa

from .enums import (
    AddAction,
    CalculationRate,
    DoneAction,
    EnvelopeShape,
    HeaderFormat,
    ParameterRate,
    SampleFormat,
)

if TYPE_CHECKING:
    import numpy

    from .osc import OscBundle, OscMessage


class Default:
    pass


class Missing:
    pass


@runtime_checkable
class SupportsOsc(Protocol):
    def to_osc(self) -> Union["OscBundle", "OscMessage"]:
        pass


@runtime_checkable
class SupportsPlot(Protocol):
    def __plot__(self) -> "numpy.ndarray":
        pass


@runtime_checkable
class SupportsRender(Protocol):
    def __render__(
        self,
        output_file_path: Optional[PathLike] = None,
        render_directory_path: Optional[PathLike] = None,
        **kwargs,
    ) -> Coroutine[None, None, Tuple[Optional[Path], int]]:
        pass


@runtime_checkable
class SupportsRenderMemo(Protocol):
    def __render_memo__(self) -> SupportsRender:
        pass


E = TypeVar("E")

_EnumLike = Optional[Union[E, SupportsInt, str, None]]

AddActionLike: TypeAlias = _EnumLike[AddAction]
DoneActionLike: TypeAlias = _EnumLike[DoneAction]
CalculationRateLike: TypeAlias = _EnumLike[CalculationRate]
FutureLike: TypeAlias = Union[concurrent.futures.Future[E], asyncio.Future[E]]
ParameterRateLike: TypeAlias = _EnumLike[ParameterRate]
RateLike: TypeAlias = _EnumLike[CalculationRate]
EnvelopeShapeLike: TypeAlias = _EnumLike[EnvelopeShape]
HeaderFormatLike: TypeAlias = _EnumLike[HeaderFormat]
SampleFormatLike: TypeAlias = _EnumLike[SampleFormat]
UGenInputMap: TypeAlias = Optional[Dict[str, Union[SupportsFloat, str, None]]]



from .core import UGen, param, ugen


@ugen(ar=True, kr=True, is_pure=True)
class AllpassC(UGen):
    """
    A cubic-interpolating allpass delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> allpass_c = supriya.ugens.AllpassC.ar(source=source)
        >>> allpass_c
        <AllpassC.ar()[0]>
    """

    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class AllpassL(UGen):
    """
    A linear interpolating allpass delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> allpass_l = supriya.ugens.AllpassL.ar(source=source)
        >>> allpass_l
        <AllpassL.ar()[0]>
    """

    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class AllpassN(UGen):
    """
    A non-interpolating allpass delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> allpass_n = supriya.ugens.AllpassN.ar(source=source)
        >>> allpass_n
        <AllpassN.ar()[0]>
    """

    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class BufAllpassC(UGen):
    """
    A buffer-based cubic-interpolating allpass delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufAllpassC.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        <BufAllpassC.ar()[0]>
    """

    buffer_id = param()
    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class BufAllpassL(UGen):
    """
    A buffer-based linear-interpolating allpass delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufAllpassL.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        <BufAllpassL.ar()[0]>
    """

    buffer_id = param()
    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class BufAllpassN(UGen):
    """
    A buffer-based non-interpolating allpass delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufAllpassN.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        <BufAllpassN.ar()[0]>
    """

    buffer_id = param()
    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class BufCombC(UGen):
    """
    A buffer-based cubic-interpolating comb delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufCombC.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        <BufCombC.ar()[0]>
    """

    buffer_id = param()
    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class BufCombL(UGen):
    """
    A buffer-based linear-interpolating comb delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufCombL.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        <BufCombL.ar()[0]>
    """

    buffer_id = param()
    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class BufCombN(UGen):
    """
    A buffer-based non-interpolating comb delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufCombN.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        <BufCombN.ar()[0]>
    """

    buffer_id = param()
    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class BufDelayC(UGen):
    """
    A buffer-based cubic-interpolating delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufDelayC.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        <BufDelayC.ar()[0]>
    """

    buffer_id = param()
    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)


@ugen(ar=True, kr=True, is_pure=True)
class BufDelayL(UGen):
    """
    A buffer-based linear-interpolating delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufDelayL.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        <BufDelayL.ar()[0]>
    """

    buffer_id = param()
    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)


@ugen(ar=True, kr=True, is_pure=True)
class BufDelayN(UGen):
    """
    A buffer-based non-interpolating delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufDelayN.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        <BufDelayN.ar()[0]>
    """

    buffer_id = param()
    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)


@ugen(ar=True, kr=True, is_pure=True)
class CombC(UGen):
    """
    A cubic-interpolating comb delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.CombC.ar(source=source)
        <CombC.ar()[0]>
    """

    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class CombL(UGen):
    """
    A linear interpolating comb delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.CombL.ar(source=source)
        <CombL.ar()[0]>
    """

    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class CombN(UGen):
    """
    A non-interpolating comb delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.CombN.ar(source=source)
        <CombN.ar()[0]>
    """

    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class DelTapRd(UGen):
    """
    A delay tap reader unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> tapin = supriya.ugens.DelTapWr.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )

    ::

        >>> tapin
        <DelTapWr.ar()[0]>

    ::

        >>> tapout = supriya.ugens.DelTapRd.ar(
        ...     buffer_id=buffer_id,
        ...     phase=tapin,
        ...     delay_time=0.1,
        ...     interpolation=True,
        ... )

    ::

        >>> tapout
        <DelTapRd.ar()[0]>
    """

    buffer_id = param()
    phase = param()
    delay_time = param(0.0)
    interpolation = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class DelTapWr(UGen):
    """
    A delay tap writer unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> tapin = supriya.ugens.DelTapWr.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )

    ::

        >>> tapin
        <DelTapWr.ar()[0]>

    ::

        >>> tapout = supriya.ugens.DelTapRd.ar(
        ...     buffer_id=buffer_id,
        ...     phase=tapin,
        ...     delay_time=0.1,
        ...     interpolation=True,
        ... )

    ::

        >>> tapout
        <DelTapRd.ar()[0]>
    """

    buffer_id = param()
    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class DelayC(UGen):
    """
    A cubic-interpolating delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.DelayC.ar(source=source)
        <DelayC.ar()[0]>
    """

    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)


@ugen(ar=True, kr=True, is_pure=True)
class DelayL(UGen):
    """
    A linear-interpolating delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.DelayL.ar(source=source)
        <DelayL.ar()[0]>
    """

    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)


@ugen(ar=True, kr=True, is_pure=True)
class DelayN(UGen):
    """
    A non-interpolating delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.DelayN.ar(source=source)
        <DelayN.ar()[0]>
    """

    source = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)


@ugen(ar=True, kr=True, is_pure=True)
class Delay1(UGen):
    """
    A one-sample delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.Delay1.ar(source=source)
        <Delay1.ar()[0]>
    """

    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class Delay2(UGen):
    """
    A two-sample delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.Delay2.ar(source=source)
        <Delay2.ar()[0]>
    """

    source = param()



from pathlib import Path

import platformdirs

from ._version import __version__, __version_info__
from .clocks import (
    AsyncClock,
    AsyncOfflineClock,
    BaseClock,
    Clock,
    ClockContext,
    OfflineClock,
    TimeUnit,
)
from .contexts import (
    AsyncServer,
    BaseServer,
    Buffer,
    BufferGroup,
    Bus,
    BusGroup,
    Context,
    Group,
    Node,
    Score,
    Server,
    Synth,
)
from .enums import (  # noqa
    AddAction,
    CalculationRate,
    DoneAction,
    HeaderFormat,
    ParameterRate,
    SampleFormat,
)
from .io import graph, play, plot, render
from .osc import OscBundle, OscCallback, OscMessage
from .patterns import Pattern
from .ugens import (
    Envelope,
    UGen,
    UGenVector,
    UGenOperable,
    SynthDef,
    SynthDefBuilder,
    synthdef,
)
from .assets.synthdefs import default
from .scsynth import Options


if not (
    output_path := Path(platformdirs.user_cache_dir("supriya", "supriya"))
).exists():
    try:
        output_path.mkdir(parents=True, exist_ok=True)
    except IOError:
        pass

assets_path = Path(__file__).parent / "assets"


__all__ = [
    "AddAction",
    "AsyncClock",
    "AsyncOfflineClock",
    "AsyncServer",
    "BaseClock",
    "BaseServer",
    "Buffer",
    "BufferGroup",
    "Bus",
    "BusGroup",
    "CalculationRate",
    "Clock",
    "ClockContext",
    "Context",
    "DoneAction",
    "Envelope",
    "Group",
    "HeaderFormat",
    "Node",
    "OfflineClock",
    "Options",
    "OscBundle",
    "OscCallback",
    "OscMessage",
    "Pattern",
    "SampleFormat",
    "Score",
    "Server",
    "Synth",
    "SynthDef",
    "SynthDefBuilder",
    "TimeUnit",
    "UGen",
    "UGenVector",
    "UGenOperable",
    "__version__",
    "__version_info__",
    "default",
    "graph",
    "output_path",
    "play",
    "plot",
    "render",
    "synthdef",
]

del Path
del platformdirs



master_doc = "index"

extensions = ["sphinx.ext.autodoc", "uqbar.sphinx.book"]

html_static_path = ["_static"]

uqbar_book_console_setup = ["import supriya"]
uqbar_book_extensions = [
    "uqbar.book.extensions.GraphExtension",
    "supriya.ext.book.PlayExtension",
    "supriya.ext.book.PlotExtension",
]
uqbar_book_strict = True
uqbar_book_use_black = True



from ...enums import ParameterRate
from ...ugens import Out, Parameter, SinOsc, SynthDef, SynthDefBuilder


def _build_test_synthdef() -> SynthDef:
    with SynthDefBuilder(
        frequency=440,
        amplitude=Parameter(value=1.0, rate=ParameterRate.AUDIO),
    ) as builder:
        sin_osc = SinOsc.ar(frequency=builder["frequency"])
        enveloped_sin = sin_osc * builder["amplitude"]
        Out.ar(bus=0, source=enveloped_sin)
    return builder.build(name="test")


def _build_test_two_voice_synthdef() -> SynthDef:
    with SynthDefBuilder(
        frequencies=(220, 440),
        amplitude=Parameter(value=1.0, rate=ParameterRate.AUDIO),
    ) as builder:
        sin_osc = SinOsc.ar(frequency=builder["frequencies"])
        enveloped_sin = sin_osc * builder["amplitude"]
        Out.ar(bus=0, source=enveloped_sin)
    return builder.build(name="test_two_voice")


test = _build_test_synthdef()
test_two_voice = _build_test_two_voice_synthdef()

__all__ = ["test", "test_two_voice"]



from typing import Any, Dict, Tuple

from ..enums import CalculationRate, DoneAction
from .basic import MulAdd
from .core import PseudoUGen, UGen, UGenOperable, UGenVector, param, ugen


@ugen(kr=True, is_pure=True)
class A2K(UGen):
    """
    An audio-rate to control-rate convert unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> a_2_k = supriya.ugens.A2K.kr(
        ...     source=source,
        ... )
        >>> a_2_k
        <A2K.kr()[0]>
    """

    source = param()


@ugen(ar=True, ir=True, kr=True, is_pure=True)
class AmpComp(UGen):
    """
    Basic psychoacoustic amplitude compensation.

    ::

        >>> amp_comp = supriya.ugens.AmpComp.ar(
        ...     exp=0.3333,
        ...     frequency=1000,
        ...     root=0,
        ... )
        >>> amp_comp
        <AmpComp.ar()[0]>
    """

    frequency = param(1000.0)
    root = param(0.0)
    exp = param(0.3333)


@ugen(ar=True, ir=True, kr=True, is_pure=True)
class AmpCompA(UGen):
    """
    Basic psychoacoustic amplitude compensation (ANSI A-weighting curve).

    ::

        >>> amp_comp_a = supriya.ugens.AmpCompA.ar(
        ...     frequency=1000,
        ...     min_amp=0.32,
        ...     root=0,
        ...     root_amp=1,
        ... )
        >>> amp_comp_a
        <AmpCompA.ar()[0]>
    """

    frequency = param(1000.0)
    root = param(0.0)
    min_amp = param(0.32)
    root_amp = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class DC(UGen):
    """
    A DC unit generator.

    ::

        >>> supriya.ugens.DC.ar(
        ...     source=0,
        ... )
        <DC.ar()[0]>

    ::

        >>> supriya.ugens.DC.ar(
        ...     source=(1, 2, 3),
        ... )
        <UGenVector([<DC.ar()[0]>, <DC.ar()[0]>, <DC.ar()[0]>])>
    """

    source = param()


@ugen(ar=True, is_pure=True)
class K2A(UGen):
    """
    A control-rate to audio-rate converter unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.kr()
        >>> k_2_a = supriya.ugens.K2A.ar(
        ...     source=source,
        ... )
        >>> k_2_a
        <K2A.ar()[0]>
    """

    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class LinExp(UGen):
    """
    A linear-to-exponential range mapper.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> lin_exp = supriya.ugens.LinExp.ar(
        ...     input_maximum=1.0,
        ...     input_minimum=-1.0,
        ...     output_maximum=22050,
        ...     output_minimum=20,
        ...     source=source,
        ... )
        >>> lin_exp
        <LinExp.ar()[0]>
    """

    source = param()
    input_minimum = param(0)
    input_maximum = param(1)
    output_minimum = param(1)
    output_maximum = param(2)


class LinLin(PseudoUGen):
    @staticmethod
    def ar(
        *,
        source,
        input_minimum=0.0,
        input_maximum=1.0,
        output_minimum=1.0,
        output_maximum=2.0,
    ) -> UGenOperable:
        scale = (output_maximum - output_minimum) / (input_maximum - input_minimum)
        offset = output_minimum - (scale * input_minimum)
        return MulAdd.new(source=source, multiplier=scale, addend=offset)

    @staticmethod
    def kr(
        *,
        source,
        input_minimum=0.0,
        input_maximum=1.0,
        output_minimum=1.0,
        output_maximum=2.0,
    ) -> UGenOperable:
        scale = (output_maximum - output_minimum) / (input_maximum - input_minimum)
        offset = output_minimum - (scale * input_minimum)
        return MulAdd.new(source=source, multiplier=scale, addend=offset)


@ugen(ar=True, kr=True, has_done_flag=True)
class Line(UGen):
    """
    A line generating unit generator.

    ::

        >>> supriya.ugens.Line.ar()
        <Line.ar()[0]>
    """

    start = param(0.0)
    stop = param(1.0)
    duration = param(1.0)
    done_action = param(DoneAction(0))

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs,
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        return calculation_rate, {
            **kwargs,
            "done_action": DoneAction.from_expr(int(kwargs["done_action"])),
        }


class Silence(PseudoUGen):
    """
    An audio-rate silence pseudo-unit generator.

    ::

        >>> supriya.ugens.Silence.ar(channel_count=2)
        <UGenVector([<DC.ar()[0]>, <DC.ar()[0]>])>
    """

    @classmethod
    def ar(cls, channel_count=1):
        from . import DC

        channel_count = int(channel_count)
        assert 0 <= channel_count
        silence = DC.ar(source=0)
        if channel_count == 1:
            return silence
        output_proxies = [silence] * channel_count
        return UGenVector(*output_proxies)


@ugen(ar=True, kr=True, has_done_flag=True)
class XLine(UGen):
    """
    An exponential line generating unit generator.

    ::

        >>> supriya.ugens.XLine.ar()
        <XLine.ar()[0]>
    """

    start = param(0.0)
    stop = param(0.0)
    duration = param(1.0)
    done_action = param(DoneAction(0))
